"""
This will eventually be its own module. But for now, it is just bundled with dfb...

rclone interfacing as a cloud API. This is intended to be like an API for a cloud 
interface and less to mimic rclone's synchronizing. As such, most method default to
always performing transfers. (via `--no-check-dest` though can change)

It is assumed that this is used by someone who is familiar with rclone and some of its
idiosyncrasies. Most methods have the ability to specify additional flags and/or
calling options. There are no guardrails so things could break! It has some built in 
methods and you can do a low-level `rclone.call()`.

It is a single file to make integrating with other codes easier.

It was started as a proof-of-concept that grew. It is still a work in progress with
some more to be added.

**TODO**: Capture failures. Especially around streaming uploads but really everywhere  
**TODO**: Better documentation of paramaters  
**TODO**: More of the example for conditional uploads which remove the always flags
**TODO**: Type annotation

NOTE: rclonerc is a better way to use rclone. Many of these features predate me using
      that API
"""

import os, sys
import subprocess
import io
import tempfile
import json
import string
import warnings
import types
import time
import logging
from functools import partialmethod, cached_property, partial

from .timestamps import timestamp_parser
from . import __version__

logger = logging.getLogger(__name__)


class RcloneCLI:
    """
    RcloneCLI object.

    Inputs:
    -------
    remote
        String remote name.

    rclone_exe ['rclone']
        Executable to call

    universal_flags ( empty tuple )
        Flags to always include. Could be thinks like config specification

    universal_env (None / Empty dict)
        Environment to awlays include. This is appended to os.environ.
        If there is something in os.environ that is *not* wanted, specify it
        with RLCONE.DELENV. ex:
            {"RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV}

    no_check_dest [True]
        On many upload operations, it is faster to include '--no-check-dest' but on
        some that can have duplicate files, such as Google Drive, set this to False
        and will use '--ignore-times' instead.

    Notes:
    -----
    - There are minimal guardrails. It is possible to specify conflicting flags.
      It is expected that you have some idea of what you're doing.

    - The general philospogy on flags is to make them keyword arguments for the most
     important or common ones.

    - Most methods have a 'flags' keyword to specify additional flags. As noted, there
      are no guard rails on them. Most also have callopts which let you control how
      rclone is called. This may be used to enable streaming the output.

    - This is designed with single files in mind. There are some methods that would be
      better off done with their own direct calls. Or this can be subclassed.

    - By design, all uploads and downloads will always happen unless wrapped with the
      'not_always' context manager.

    - Anytime an RcloneFile object is passed as a source, it's path is used. This enables
      remote-to-remote transfers. Note that the flag and environment of the main object
      is used. This means that, for example, you cannot have remotes defined in their own
      config files.


    Useful Flags:
    -------------
    It may be useful to set '--links', '--copy-links', or '--skip-links' but they are not
    otherwise set

    """

    NO_TRAVERSE = 50  # WAG of a cutoff for multiple uploads
    BLOCKSIZE = 1024  # 1kb. For reading inputs

    # Sentinels. Made with os.urandom(8)
    DELENV = "**DELENV**"  # Remove from environment
    NOFLAG = "XTfRDcsimw"  # Remove from call

    def __init__(
        self,
        remote,
        *,
        rclone_exe="rclone",
        universal_flags=None,
        universal_env=None,
        no_check_dest=True,
    ):
        self.remote = remote
        self.rclone_exe = rclone_exe
        self.uflags = [str(flag) for flag in _flagify(universal_flags)]

        # Save this for presenting
        self.universal_env = universal_env = {
            str(k): str(v) for k, v in _dictify(universal_env).items()
        }

        # But this is for saved for real use
        self.uenv = os.environ.copy()
        self.uenv.update(universal_env)
        for key, val in universal_env.items():
            if val == self.DELENV:
                self.uenv.pop(key, None)
                universal_env[key] = "**UNSET**"

        self.always_flag = "--no-check-dest" if no_check_dest else "--ignore-times"

        self._capture = False

    # Main Actions. All also contains flag and callopts arguments for more control

    def listremote(
        self,
        subdir="",
        *,
        filters=None,
        filter_flags=None,
        fast_list=False,
        mimetype=False,
        modtime=True,
        hashes=False,
        hashtypes=False,
        metadata=False,
        maxdepth=None,
        only=None,
        epoch_time=False,
        # pipe=False,
        flags=None,
        callopts=None,
    ):
        """
        List the remote with lsjson.

        Some common flags are built in options for convenience but any additional
        can be specified manually. See https://rclone.org/commands/rclone_lsjson/

        The only processing is to convert ModTime to Python datetime objects

        Inputs:
        -------
        subdir [empty]
            Subdir to list. Note that paths will be relative TO THIS DIR.
            You can use os.path.join on the output if needed.

            Can also be a specific file but will still return an iterator.
                >>> fileinfo = list(rclone.listremote(subdir='myfile.ext'))[0]

        filters [empty]
            RcloneCLI filters as items. See also filter_flags

        filter_flags [empty]
            Additional filtering flags such as
                --include, --exclude, --include-from, --exclude-from,
                --filter, --filter-from, --files-from,
                --one-file-system, --exclude-if-present

            This is effectively the same as flags but more explicit.

        fast_list [False]
            Whether to add --fast-list. Or specify 'auto' to be True if
            the remote supports it.

        mimetype [False]
            If False, will call with --no-mimetype

        modtime [True]
            If False, will call with --no-modtime. This can be faster on remotes
            that require additional requests for it. Note that it still may or may not
            include a ModTime. ModTimes are converted to datetime objects unless
            epoch_time=True.

        hashes [False]
            Compute hashes (adds '--hash')

        hashtypes [None]
            If specified and hashes=True, will use these types. String of one or list
            of types

        metadata [False]
            If True, adds --metadata

        maxdepth [None]
            Adds --max-depth flags.

        only
            Specify as {None [default],'files','dirs'}.

        epoch_time [False]
            If True, returns a Unix Epoch float. Otherwise, returns a time zone aware
            datetime object

        flags [empty]
            Additional rclone flags. No guardrails or error checking!

        callopts [empty]
            Additional options passed to call. No guardrails or error checking!

        Returns:
        -------
        Iterator on file items

        """
        """
        pipe [False]
            Whether to write output to an intermediate file.
        """
        pipe = True  # TODO...

        subdir = subdir or ""  # Convert None to ""

        cmd = ["lsjson", RcloneCLI.pathjoin(self.remote, subdir), "--recursive"]
        if fast_list == "auto":
            fast_list = rclone.features.get("ListR", False)
        if fast_list:
            cmd.append("--fast-list")

        for ff in _flagify(filters):
            cmd.append("--filter")
            cmd.append(ff)
        cmd += _flagify(filter_flags)

        if not mimetype:
            cmd.append("--no-mimetype")
        if not modtime:
            cmd.append("--no-modtime")
        if hashes:
            cmd.append("--hash"),
            if hashtypes:  # Only if hashes
                hashtypes = _flagify(hashtypes)
                for hashtype in hashtypes:
                    cmd.extend(["--hash-type", hashtype])
        if metadata:
            cmd.append("--metadata")
        if maxdepth:
            cmd.extend(["--max-depth", str(maxdepth)])

        if only not in {None, "dirs", "files"}:
            raise ValueError("'only' must be one of {None,'dirs','files'}")
        if only:
            cmd.append(f"--{only}-only")

        cmd += _flagify(flags)

        # Long directories can be problematic so use a tempfile so as to not exhaust the
        # buffer. We also want to parse this lazily since it could be long. This isn't
        # perfect since we wait for the entire listing to finish then read line by line
        # but it avoids accidentally deadlocking.
        res = self.call(cmd, pipe=pipe, stream=True, **_dictify(callopts))

        # Special case for '--stat' whether user specified or from iteminfo.
        # RcloneCLI doesn't do the one-line-per-response with this call.
        # A bit of a hack but we want it to pass through the for loop processing still.
        if "--stat" in cmd:
            lines = []
            for oe, line in res:
                if oe != "stdout":
                    continue
                lines.append(line)
            item = json.loads(b"".join(lines))
            res = [("stdout", json.dumps(item).encode())]

        for oe, line in res:
            if oe != "stdout":
                logger.debug(f"stdout: {line}")
                continue

            # lsjson returns one entry per line. And always UTF8
            line = line.decode("utf8")
            line = line.strip().rstrip(",").strip()

            if line == "[" or line == "]":  # start or end line
                continue

            line = json.loads(line)

            # Never understood why rclone gives us this...
            line.pop("Name", None)

            if "ModTime" in line:  # Do regardless of modtime setting
                line["ModTime"] = timestamp_parser(line["ModTime"], epoch=epoch_time)
            yield line

    ls = listremote

    def iteminfo(self, remoteitem, **kwargs):
        """
        List a single item. This is a convenience function around listremote that just
        returns the first item.

        Inputs:
        -------
        remoteitem
            Remote item of interest. If empty string, will be the root directory of
            the remote

        **kwargs
            Passed to listremote

        """
        kwargs["flags"] = _flagify(kwargs.get("flags", None)) + ["--stat"]
        kwargs["subdir"] = remoteitem

        info = list(self.listremote(**kwargs))
        return info[0]

    def upload(self, local, destdir="", *, flags=None, callopts=None):
        """
        Upload a single file 'local' to the 'destdir' keeping the name.

        Inputs:
        -------
        local
            Local file to upload

        destdir ['']
            Where on the remote to upload

        flags [empty]
            Additional rclone flags. No guardrails or error checking!

        callopts [empty]
            Additional options passed to call. No guardrails or error checking!
            Useful ones are 'return_proc' and 'stream'

        Returns:
        -------
        results of .call(). Usually stdout,stderr unless callpopts were changed
        """
        cmd = [
            "copy",
            local,
            RcloneCLI.pathjoin(self.remote, destdir),
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always upload
        ] + _flagify(flags)

        return self.call(cmd, **_dictify(callopts))

    def uploadto(self, local, destfile, *, flags=None, callopts=None):
        """
        Upload a single file 'local' to the 'destfile' with any name

        Inputs:
        -------
        local
            Local file to upload

        destfile ['']
            Destination file name

        flags [empty]
            Additional rclone flags. No guardrails or error checking!

        callopts [empty]
            Additional options passed to call. No guardrails or error checking!
            Useful ones are 'return_proc' and 'stream'

        Returns:
        -------
        results of .call(). Usually stdout,stderr unless callpopts were changed
        """
        cmd = [
            "copyto",
            local,
            RcloneCLI.pathjoin(self.remote, destfile),
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always upload
        ] + _flagify(flags)

        return self.call(cmd, **_dictify(callopts))

    def uploadmany(
        self,
        files,
        destdir="",
        upload_from=".",
        *,
        flags=None,
        callopts=None,
    ):
        """
        Upload many files from 'upload_from' location. Can combine with the not_always
        context manager.

        Inputs:
        ------
        files
            List of file paths. They should be relative to 'upload_from'

        destdir ['']
            Where on the remote to upload

        upload_from ['.']
            The local "remote"

        flags [empty]
            Additional rclone flags. No guardrails or error checking!

        callopts [empty]
            Additional options passed to call. No guardrails or error checking!
            Useful ones are 'return_proc' and 'stream'

        Returns:
        -------
        results of .call(). Usually stdout,stderr unless callpopts were changed
        """
        if isinstance(files, str):
            files = [files]
        files = list(files)

        with tempfile.NamedTemporaryFile(delete=False, mode="wt") as fp:
            for file in files:
                if file.startswith(".."):
                    msg = f"Will not upload files above specified location. {file!r}"
                    warnings.warn(msg)
                print(file, file=fp)

        cmd = [
            "copy",
            upload_from,
            RcloneCLI.pathjoin(self.remote, destdir),
            self.always_flag,  # Always upload
            "--files-from",
            fp.name,
        ] + _flagify(flags)

        # I am not sure about this one. I think that if you are always uploading, it
        # doesn't matter. Or maybe just for no-check-dest. In which case, this is
        # probably implied anyway.
        if len(files) < self.NO_TRAVERSE:
            cmd += ["--no-traverse"]

        return self.call(cmd, **_dictify(callopts))

    def write(
        self,
        data,
        destfile,
        *,
        size=None,
        fallback=False,
        flags=None,
        callopts=None,
    ):
        """
        Upload data to a file

        Inputs:
        -------
        data
            Data to upload. If a string, it will be UTF8 encoded to bytes.
            If it a file object, it will be read in RcloneCLI.BLOCKSIZE blocks
            and uploaded. If a types.GeneratorType, will iterates and sent
            to rclone

        destfile ['']
            Destination file name

        size [None]
            Specify the size of it is known. This will enable more backends to
            directly support uplaods. If given a string or bytes, this will
            be automatically calculated

        fallback [False]
            Fallback to writing a temp file and uploading. Can also set to 'auto'

        flags [empty]
            Additional rclone flags. No guardrails or error checking!

        callopts [empty]
            Additional options passed to call. No guardrails or error checking!
            Useful ones are 'return_proc' and 'stream'

        Returns:
        -------
        results of .call(). Usually stdout,stderr unless callpopts were changed

        """
        if fallback == "auto":
            opts = dict(size=size, flags=flags, callopts=callopts)
            try:
                return self.write(data, destfile, fallback=False, **opts)
            except:  # Allow it to be anything!
                logger.error("write failed. Trying fallback")
                return self.write(data, destfile, fallback=True, **opts)

        ###########################

        if isinstance(data, str):
            data = data.encode("utf8")

        if fallback:
            with tempfile.NamedTemporaryFile() as fp:
                if hasattr(data, "read"):
                    while block := data.read(self.BLOCKSIZE):
                        fp.write(block)
                elif isinstance(data, types.GeneratorType):
                    for block in data:
                        fp.write(block)
                elif isinstance(data, bytes):
                    fp.write(data)
                else:
                    raise TypeError("Could not handle input")
                fp.flush()
                return self.uploadto(fp.name, destfile, flags=flags, callopts=callopts)

        if isinstance(data, bytes):
            size = len(data)

        cmd = [
            "rcat",
            RcloneCLI.pathjoin(self.remote, destfile),
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always upload
        ] + _flagify(flags)
        if size:
            cmd.extend(["--size", str(size)])

        return self.call(cmd, stdin=data, **_dictify(callopts))

    def download(self, remotefile, destdir, *, flags=None, callopts=None):
        """downloads a remotefile file to the destdir"""
        cmd = [
            "copy",
            RcloneCLI.pathjoin(self.remote, remotefile),
            destdir,
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always download
        ] + _flagify(flags)

        return self.call(cmd, **_dictify(callopts))

    def downloadto(self, remotefile, destfile, *, flags=None, callopts=None):
        """Upload a local file to the destfile"""
        cmd = [
            "copyto",
            RcloneCLI.pathjoin(self.remote, remotefile),
            destfile,
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always download
        ] + _flagify(flags)

        return self.call(cmd, **_dictify(callopts))

    def read(
        self,
        remotefile,
        *,
        offset=None,
        count=None,
        fileobject=False,
        flags=None,
        callopts=None,
    ):
        """
        Return the bytes of the file or a fileobject. Note that changing the
        pipe settings with callopts may require you to seek() the file.
        """
        cmd = ["cat", RcloneCLI.pathjoin(self.remote, remotefile)]
        if offset:
            cmd.extend(["--offset", str(offset)])
        if count:
            cmd.extend(["--count", str(count)])

        cmd += _flagify(flags)

        kw = _dictify(callopts)

        if fileobject:
            kw["return_proc"] = True
            res = self.call(cmd, **kw).stdout
        else:
            kw["return_proc"] = False
            kw["pipe"] = True
            res = self.call(cmd, **kw)

        if isinstance(res, tuple):  # default
            return res[0]
        return res  # Maybe a callopts set it to the proc

    def delete(
        self,
        remotefile,
        *,
        filters=None,
        rmdirs=False,
        flags=None,
        callopts=None,
    ):
        """
        Delete a remote path, potentially including directories.
        """

        cmd = ["delete", RcloneCLI.pathjoin(self.remote, remotefile)] + _flagify(flags)

        if rmdirs:
            cmd.append("--rmdirs")

        for ff in _flagify(filters):
            cmd.append("--filter")
            cmd.append(ff)

        return self.call(cmd, **_dictify(callopts))

    def open(
        self,
        remotefile,
        mode="rb",
        buffer_size=8 * 1024 * 1024,
        flags=None,
        callopts=None,
    ):
        """
        Return a buffered, seekable, file-like object of 'remotefile',
        that is buffered with 'buff_size'. Default buffer_size is 8mb or
        8*1024*1024 = 8388608 bytes

        Warning though, if called with .read(), will use a *much* smaller
        buffer. It will work but can be extremely slow!

        """
        if "w" in mode:
            raise io.UnsupportedOperation("Cannot write to stream")
        raw = _RawRcloneFileObj(remotefile, self, flags=flags, callopts=callopts)
        fp = io.BufferedReader(raw, buffer_size=buffer_size)
        if "b" in mode:
            return fp
        return io.TextIOWrapper(fp)

    def _movecopy(self, remotesrc, remotedst, *, _cmd, flags=None, callopts=None):
        cmd = [
            _cmd,
            RcloneCLI.pathjoin(self.remote, remotesrc),
            RcloneCLI.pathjoin(self.remote, remotedst),
            "--no-traverse",  # Single file. This is better
            self.always_flag,  # Always
        ] + _flagify(flags)

        return self.call(cmd, **_dictify(callopts))

    move = partialmethod(_movecopy, _cmd="move")
    copy = partialmethod(_movecopy, _cmd="copy")
    moveto = partialmethod(_movecopy, _cmd="moveto")
    copyto = partialmethod(_movecopy, _cmd="copyto")

    @cached_property
    def version(self):
        cmd = ["version"]
        res, _ = self.call(cmd)
        return res

    @cached_property
    def version_dict(self):
        cmd = ["rc", "--loopback", "core/version"]
        res, _ = self.call(cmd)
        return json.loads(res)

    @cached_property
    def backend_features(self):
        cmd = ["backend", "features", self.remote]
        res, _ = self.call(cmd)
        return json.loads(res)

    @property
    def features(self):
        return self.backend_features.get("Features", {})

    @cached_property
    def config_paths(self):
        out, _ = self.call(["config", "paths"])
        paths = {}
        for line in out.decode().splitlines():
            line = line.strip()
            if not line:
                continue
            key, val = line.split(":", 2)
            paths[key.strip()] = val.strip()
        return paths

    def call(
        self,
        cmd,
        *,
        stdin=None,
        pipe=True,
        stream=False,
        allow_error=False,
        return_proc=False,
    ):
        """
        Call rclone and get results, or Popen object, or iterator over lines

        Options:
        -------
        stdin [None]
            What to send to rclone. If bytes, will send directly. If a file
            object, will send in RcloneCLI.BLOCKSIZE (default 1kb) chunks. If a
            types.GeneratorType, will iterate and sent to rclone

        pipe [True]
            Whether or not to use subprocess.PIPE (buffered in memory) or write
            to a temp file. Even if false, will still work with stream.

        stream: [False]
            Return a generator of (stdin -or- stdout,line) pairs that can be printed
            to the user right away. Sets return_proc=False but will still works with
            either pipe setting. If pipe is False, raises an error.

            Note that streaming output will not start until stdin has been fully written.
            It is possible to work around this limitation but not worth the added
            complexity.

        allow_error [False]
            If True, does not raise an error on failure. Ignored for return_proc
            but not for stream.

        return_proc: [False]
            If True, will return the Popen object rather than call communicate.
            It will always have stdin,stdout variables defined even but with buffer=False,
            they will need to be seeked to the start or should be open as a new reader.



        """
        finalcmd = [self.rclone_exe] + self.uflags + cmd
        finalcmd = [c for c in finalcmd if c != self.NOFLAG]
        finalcmd = [c if not isinstance(c, RcloneFile) else c.path for c in finalcmd]

        logger.debug(
            f"rclone call {str(finalcmd)} with additional "
            f"environ {json.dumps(self.universal_env)}"
        )

        if pipe:
            stdout = subprocess.PIPE
            stderr = subprocess.PIPE
        else:
            if stream:
                raise ValueError("Cannot stream without pipe=True")
            stdout = tempfile.NamedTemporaryFile(delete=False)
            stderr = tempfile.NamedTemporaryFile(delete=False)

        # Include this here so we can replace it in either a subclass or monkey-patch
        proc = self._Popen(
            finalcmd,
            stdout=stdout,
            stderr=stderr,
            stdin=subprocess.PIPE,
            env=self.uenv,
        )
        if not proc:
            return

        if not pipe:
            proc.stdout, proc.stderr = stdout, stderr

        if hasattr(stdin, "read"):
            # Do it this ways instead of setting this as stdin
            # since it doesn't require stdin to be a real file.
            while block := stdin.read(self.BLOCKSIZE):
                proc.stdin.write(block)
        elif isinstance(stdin, types.GeneratorType):
            for block in stdin:
                proc.stdin.write(block)
        elif stdin:
            # Do it here rather than in communicate in case we are returning ot
            proc.stdin.write(stdin)

        if stream:
            # Technically, the stream is still imperfect since it won't start streaming
            # until we've written all of stdin. It could be done via threads but the
            # added complexity is not worth it. And it will be *much* harder to ensure
            # no deadlocks. So we write all of stdin before we start reading stdout.
            # This is noted above
            proc.stdin.close()

            return popen_streamer(proc, allow_error=allow_error)

        if return_proc:
            proc.stdin.close()
            return proc

        out, err = proc.communicate()  # Also does proc.wait()

        if not allow_error:
            check_returncode(proc, output=out, stderr=err)

        if not pipe:
            return stdout, stderr

        return out, err

    def _Popen(self, *args, **kwargs):
        return subprocess.Popen(*args, close_fds=True, **kwargs)

    # def logger.debug(self, *args, **kwargs):
    #         """Debug. Can subclass as needed"""
    #         return logger.logger.debug(*args, **kwargs)

    @staticmethod
    def pathjoin(*args, local_root=False):
        """
        This is like os.path.join but does some rclone-specific things because
        there could be a ':' in the first part.

        The second argument could be '/file', or 'file' and the first could have a colon.
            RcloneCLI.pathjoin('a','b')   # a/b
            RcloneCLI.pathjoin('a:','b')  # a:b
            RcloneCLI.pathjoin('a:','/b') # a:/b # Note that these are unlike os.path.join
            RcloneCLI.pathjoin('a','/b')  # a/b

        """
        args = [str(a) for a in args]  # Pathlib

        if len(args) <= 1:
            return "".pathjoin(args)

        root, first, rest = args[0], args[1], args[2:]

        if root.endswith("/"):
            root = root[:-1]

        if root.endswith(":") or first.startswith("/"):
            path = root + first
        else:
            path = f"{root}/{first}"

        path = os.path.join(path, *rest)
        return path

    @property
    def not_always(self, **kwargs):
        """
        Context manager to disable unconditional uploads. Useful if you
        do not want to force an upload if rclone thinks they are the same.

        Note that you may want to add additional flags to control it.

        Examples:

        Normal. Will always upload

            >>> rclone.upload('existing_file.ext')

        Use default checks:

            >>> with rclone.not_always:
            >>>     rclone.upload('existing_file.ext')

        Use '--checksum' mode:

            >>> with rclone.not_always:
            >>>     rclone.upload('existing_file.ext',flags=['--checksum'])
        """
        return enable_check_dest(self, **kwargs)

    enable_check_dest = not_always

    def capture(self, execute=True, save_results=False, save_stdin=False):
        """
        Context Manager to capture the subprocess command and optionally, the results.

        Can also tell it not to execute but this may cause issues downstream.

        Example:

            IN: rclone = RcloneCLI('myremote:')
                with rclone.capture(save_results=True) as capture:
                    rclone.write('test data','data.txt')
                print(capture.command_history[0])
                print(capture.command_results[0])

           OUT: ['rclone', 'rcat', 'myremote:data.txt',
                 '--no-traverse', '--no-check-dest', '--size', '9']

                 (b'', b'') # No output but captured

        Note that you can use rclone.universal_env to get the *specified* environment
        or rclone.uenv to get the *actual* used environment

        Again, execute=False may cause downstream failure.

        If using save_stdin, it saves bytes or strings. If stdin is specified as
        something else, it will be b'**not bytes**'. Otherwise, it's None.

        Warning: Not thread-safe. May be fixed in the future
        """
        return capture(
            self, execute=execute, save_results=save_results, save_stdin=save_stdin
        )

    def __repr__(self):
        return f"RcloneCLI(remote={self.remote!r})"

    def __truediv__(self, new):
        return RcloneFile(self, new)


### RcloneCLI Objects that can be used for source
class RcloneFile:
    """
    Object representing many rclone operations. Note that move and moveto will
    update in place. Copy/Copyto does not
    """

    def __init__(self, rcloneobj, remoteitem=""):
        self.rclone = rcloneobj
        self.remoteitem = remoteitem
        self.path = RcloneCLI.pathjoin(self.rclone.remote, self.remoteitem)
        self.fs_remote = self.rclone.remote, self.remoteitem

    def __truediv__(self, new):
        # Need to decide if using RcloneCLI.pathjoin or os.path.join
        if self.remoteitem:
            return RcloneFile(self.rclone, os.path.join(self.remoteitem, new))
        return RcloneFile(self.rclone, new)

    def __repr__(self):
        return f"RcloneFile({self.path})"

    def __str__(self):
        return f"{self.path}"

    def _partial_remoteitem(self, name, *args, **kwargs):
        """Make self.remoteitem the first argument"""
        func = getattr(self.rclone, name)
        return func(self.remoteitem, *args, **kwargs)

    def _partial_remoteitemR(self, name, *args, **kwargs):
        """Rightsided"""
        func = getattr(self.rclone, name)
        return func(*args, self.remoteitem, **kwargs)

    # Right sided partial
    upload = partialmethod(_partial_remoteitemR, "uploadto")
    write = partialmethod(_partial_remoteitemR, "write")

    # Left sided partial
    download = partialmethod(_partial_remoteitem, "download")
    downloadto = partialmethod(_partial_remoteitem, "downloadto")
    read = partialmethod(_partial_remoteitem, "read")
    delete = partialmethod(_partial_remoteitem, "delete")

    copy = partialmethod(_partial_remoteitem, "copy")
    moveto = partialmethod(_partial_remoteitem, "moveto")
    copyto = partialmethod(_partial_remoteitem, "copyto")
    open = partialmethod(_partial_remoteitem, "open")
    info = iteminfo = partialmethod(_partial_remoteitem, "iteminfo")

    def move(self, remotedst, **kwargs):
        r = self.rclone.move(self.remoteitem, remotedst, **kwargs)
        self.remoteitem = os.path.join(remotedst, os.path.basename(self.remoteitem))
        return r

    def moveto(self, remotedst, **kwargs):
        r = self.rclone.moveto(self.remoteitem, remotedst, **kwargs)
        self.remoteitem = remotedst
        return r


### File Object
class _RawRcloneFileObj(io.RawIOBase):
    """Do not call this without a buffer wrapping it!"""

    def __init__(self, remotefile, rclone, *, flags=None, callopts=None):
        self.rclone = rclone
        self.remotefile = remotefile
        self.offset = 0
        self.flags = flags
        self.callopts = callopts
        self.maxsize = None

    def seekable(self):
        return True

    def seek(self, offset, whence=io.SEEK_SET):
        if whence == io.SEEK_SET:
            self.offset = offset
        elif whence == io.SEEK_CUR:
            self.offset += offset
        else:  # whence = io.SEEK_END
            raise io.UnsupportedOperation()
        return self.offset

    def tell(self):
        return self.offset

    def readable(self):
        return True

    def writable(self):
        return False

    def readinto(self, b):
        if self.maxsize and self.offset >= self.maxsize:
            return 0
        N = len(b)
        chunk = self.rclone.read(
            self.remotefile,
            offset=self.offset,
            count=N,
            flags=self.flags,
            callopts=self.callopts,
        )
        n = len(chunk)
        self.offset += n
        if n != N:  # We know we hit the end
            self.maxsize = self.offset
        b[:n] = chunk
        return n


### Utilities


def popen_streamer(proc, allow_error=False):
    """
    Takes a subprocess.Popen object and yields stdout and stderr
    simultaneously

    The proc must be opened with
        stdout=subprocess.PIPE,stderr=subprocess.PIPE

    Note that if the underlying process doesn't flush the output,
    it may not happen as expected.

    if PIPE is False, tells this function that subprocess if being
    written to files and will read them with a poll time of 1/100th
    of a second.

    Yields:
        ('stdout' or 'stderr', line)
    """
    from threading import Thread
    from queue import Queue

    Q = Queue()

    def _reader(oe):
        file = getattr(proc, oe)
        with file:
            for line in file:
                Q.put((oe, line))

        Q.put((oe, None))  # Done

    outthread = Thread(target=_reader, args=("stdout",))
    errthread = Thread(target=_reader, args=("stderr",))
    outthread.start(), errthread.start()

    c = 0
    while c < 2:
        oe, line = Q.get()
        if line is None:
            c += 1
            Q.task_done()
            continue
        yield oe, line
        Q.task_done()

    proc.wait()  # Should be done executing already
    if not allow_error:
        check_returncode(proc)

    Q.join()
    outthread.join()
    errthread.join()


def _flagify(flags):
    flags = flags or []
    if isinstance(flags, str):
        flags = [flags]
    return list(flags)


def _dictify(mydict):
    if not mydict:
        return {}
    if isinstance(mydict, (list, tuple)):
        return dict(mydict)
    return mydict


class enable_check_dest:
    """
    Remove --no-check-dest or --ignore-times
    """

    def __init__(self, rclone):
        self.rclone = rclone

    def __enter__(self):
        self.af = self.rclone.always_flag
        self.rclone.always_flag = self.rclone.NOFLAG

    def __exit__(self, exc_type, exc_value, traceback):
        self.rclone.always_flag = self.af


class capture:
    """
    Capture rclone calls and optionally return False to no continue.

    WARNING: setting execute=False may cause errors downstream!

    Variables:
    ----------
    command_history
    command_results (iff save_results and execute)
    command_stdin (iff save_stdin and stdin is bytes or string)

    """

    def __init__(self, rclone, execute=True, save_results=False, save_stdin=False):
        self.rclone = rclone
        self.execute = execute
        self.save_results = save_results
        self.save_stdin = save_stdin
        self.command_history = []
        self.command_stdin = []
        self.command_results = []

    def __enter__(self):
        self._oldPopen, self.rclone._Popen = self.rclone._Popen, self._Popen
        self._oldcall, self.rclone.call = self.rclone.call, self._call
        self.rclone._capture = True
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.rclone._Popen = self._oldPopen
        self.rclone.call = self._oldcall
        self.rclone._capture = False

    def _Popen(self, cmd, *args, **kwargs):
        self.command_history.append(cmd)
        if self.execute:
            return subprocess.Popen(cmd, *args, **kwargs)
        return False

    def _call(self, *args, **kwargs):
        if self.save_stdin:
            stdin = kwargs.get("stdin", None)
            if isinstance(stdin, (str, bytes)):
                self.command_stdin.append(stdin)
            elif stdin:
                self.command_stdin.append(b"**not bytes**")
            else:
                self.command_stdin.append(None)
        else:
            self.command_stdin.append(None)

        res = self._oldcall(*args, **kwargs)
        if self.save_results:
            self.command_results.append(res)
        return res

    def shell_script(self, export=True, cd=True):
        """
        Return a string that represents the shell script to do the same as the results.

        Options:
        -------
        export [True]
            Include  `export` statements of the rclone environment

        cd [True]
            Include a `cd` command to the current directory
        """
        import shlex

        out = []
        if cd:
            cmd = ["cd", os.path.abspath(os.getcwd())]
            out.append(shlex.join(cmd))
        if export:
            for key, value in self.rclone.universal_env.items():
                if value in ("**UNSET**", self.rclone.DELENV):
                    out.append(f"unset {key}")
                    continue
                out.append(f"export {key}={shlex.quote(value)}")
        for stdin, cmd in zip(self.command_stdin, self.command_history):
            cmd = shlex.join(cmd)
            if stdin:
                if isinstance(stdin, bytes):
                    try:
                        stdin = stdin.decode("utf8")
                    except:
                        stdin = "**not UTF-8**"
                        out.append("# WARNING -- Could not understand stdin")
                if stdin == "**not bytes**":
                    out.append(
                        "# WARNING -- stdin specified and not bytes. May not work"
                    )
                cmd = shlex.join(["echo", stdin]) + f" | {cmd}"
            out.append(cmd)

        return "\n".join(out)


def check_returncode(proc, output=None, stderr=None):
    """Raise CalledProcessError if the exit code is non-zero."""

    if proc.returncode:
        raise subprocess.CalledProcessError(
            proc.returncode, proc.args, output=output, stderr=stderr
        )



"""
Database of the destination. Includes the tools and methods to refresh it.
"""

import os
from pathlib import Path
import sqlite3
import time
import io
import json
import logging
import string
import shutil
import gzip as gz
from functools import partialmethod
from textwrap import dedent, indent

from . import __version__, nowfun, rpath2apath, apath2rpath
from .utils import (
    time2all,
    MyRow,
    star,
    listify,
    smart_open,
    randstr,
    smart_splitext,
    NoTimestampInNameError,
)
from .timestamps import timestamp_parser
from . import rclonerc
from .rclonerc import IGNORED_FILE_DATA, rcpathjoin
from .threadmapper import thread_map_unordered as tmap


logger = logging.getLogger(__name__)


def sqldebug(sql):
    return
    sql = "\n".join(line for line in sql.split("\n") if line.strip())
    sql = dedent(sql).rstrip()
    logger.info(f"SQL >>>>>>>>>>>>>>> DSTDB\n{sql}\n<<<<<<<<<<<<<<<")


class DFBDST:
    """
    Main database object for the destination
    """

    COLS = (
        ("rpath", "TEXT NOT NULL"),  # Full path to the real file
        ("apath", "TEXT NOT NULL"),  # Full path to aparent name
        ("timestamp", "INTEGER NOT NULL"),
        ("size", "INTEGER"),
        ("mtime", "REAL"),
        ("checksum", "TEXT"),
        ("isref", "INTEGER"),  # 0: not ref, 1: ref, 2: ref not updated
        ("ref_rpath", "TEXT"),
        ("dstinfo", "INTEGER"),  # Information is from the dest, not source
        ("remain", "TEXT"),
    )

    def __init__(self, config):
        self.config = config
        self.dst_rclone = dst_rclone = config._config["dst_rclone"]
        self.dbcache_dir = config.dbcache_dir

        self.snap_file = (
            self.config.snap_cache_dir
            / self.config.now.obj.strftime("%Y/%m")
            / f"{self.config.now.dt}Z.jsonl"
        )
        self.snap_file.parent.mkdir(exist_ok=True, parents=True)

        dbpath = self.dbcache_dir / f"{config.config_id}.db"
        dbpath.parent.mkdir(exist_ok=True, parents=True)

        self.dbpath = dbpath

        self.init()

    def db(self):
        db = sqlite3.connect(self.dbpath, check_same_thread=True)
        db.row_factory = MyRow
        db.set_trace_callback(sqldebug)
        return db

    def init(self):
        # We will only write to the DB in the main thread but will
        # read in many
        items = ",".join((" ".join(row)) for row in self.COLS)
        db = self.db()

        try:
            with db:
                db.execute("PRAGMA journal_mode = wal")
        except sqlite3.OperationalError:  # pragma: no cover
            pass

        # test:
        try:
            with db:
                r = db.execute(
                    """
                    SELECT * FROM kv 
                    WHERE key = ? OR key = ?
                    ORDER BY key""",
                    ("created", "version"),
                ).fetchall()
                if len(r) == 2:  # Note it is ORDER BY so the order wont change
                    created, version = [i["val"] for i in r]
                    logger.debug(f"dstdb exists. {created = } {version = }")
                    return
        except:
            logger.debug("Recreate dstdb")

        with db:
            db.execute(
                f"""
                CREATE TABLE IF NOT EXISTS
                items(
                    {items},
                    PRIMARY KEY (apath, timestamp)
                )"""
            )

            db.execute(
                """
                CREATE TABLE IF NOT EXISTS kv(
                    key TEXT PRIMARY KEY,
                    val BLOB
                )"""
            )

            db.execute(
                """
                INSERT OR IGNORE INTO kv VALUES (?,?)
                """,
                ("created", self.config.now.obj.isoformat()),
            )
            db.execute(
                """
                INSERT OR IGNORE INTO kv VALUES (?,?)
                """,
                ("version", __version__),
            )
        db.commit()
        db.close()

    def reset(self, stats=None, *, use_snapshots):
        if self.config.disable_refresh:
            logger.error("Refresh not allowed due to 'disable_refresh = True")
            logger.error("Run with `--override 'disable_refresh = False'` to override")
            raise ValueError("Refresh Disabled")

        self.dbpath.unlink()
        self.init()

        files = list(self._relist(stats=stats))  # Need them all to get snapshots
        logger.info(
            f"Found {len(files)} at dest "
            f"with {sum(f.get('isref',0) == 2 for f in files)} reference(s)"
        )

        if use_snapshots:
            self._load_snapshots()
            files = (self._apply_snapshot_file(file) for file in files)

        db = self.db()
        for ii, file in enumerate(files):
            db.execute(
                f"INSERT INTO items VALUES ({','.join('?' for _ in DFBDST.COLS)})",
                DFBDST.dict2fullrow(file),
            )
            if ii % 100 == 0:  # Frequent commits but not every time.
                db.commit()
        db.commit()
        db.close()

        # Update those with isref = 2. Do this after full listing
        self._update_references()

    def _relist(self, stats=None):
        self._snapshot_list = []

        config = self.config
        flags = config.dst_list_rclone_flags

        files = config.dst_rclone.listremote(
            mimetype=False,
            # Notice modtime and hashes are only if needed and not using get_modtime
            # or get_hashes
            modtime="mtime" in (config.dst_compare, config.dst_renames),
            hashes="hash" in (config.dst_compare, config.dst_renames),
            hashtypes=config.hash_type,
            # metadata=... # Set in universal_flags#
            only="files",
            epoch_time=True,
            flags=flags,  # Will include fast-list if needed
            #             pipe=False,
            filters=["- **/.swap.*", "- /.dfb/**"],
        )

        t0 = time.time()
        c = 0
        for file in files:
            try:
                apath, ts, flag = rpath2apath(file["Path"])
            except (ValueError, NoTimestampInNameError, IndexError):
                logger.error(f"Could not find timestamp for {file['Path']}. Ignoring")
                continue
            c += 1

            size = file.pop("Size")
            new = {
                "rpath": file.pop("Path"),
                "apath": apath,
                "timestamp": ts,
                "size": size if flag != "D" else -1,
                "mtime": file.pop("ModTime", None),
                "isref": 2 if flag == "R" else 0,  # 2 means not yet updated. Later
                "dstinfo": True,
            }
            if hashes := file.pop("Hashes", None):
                new["checksum"] = hashes

            # Update with everything else
            for k, v in file.items():
                if k in IGNORED_FILE_DATA:
                    continue
                new[k] = v

            if stats and (time.time() - t0) >= stats:  # TODO TEST
                logger.info(f"Destination Listing Status: {c} items")
                t0 = time.time()

            yield new

    def _update_references(self):
        db = self.db()
        with db:
            files = db.execute("""SELECT * FROM items WHERE isref = 2""")
            files = files.fetchall()

        if not files:
            return

        logger.info(f"Need to fetch {len(files)} references")

        # Multi-thread reading from the remote to get the new rpath
        # and reading from the DB to get the info
        rc = self.config.rc
        rc.start()

        def _get_referent(file):
            refferer = file["rpath"]
            referent = rc.read((self.config.dst, refferer)).decode()

            # Handle different versions here
            try:
                referent = json.loads(referent)
            except json.JSONDecodeError:
                logger.debug(f"Reading reference. Assuming V1")
                referent = {"ver": 1, "path": referent}

            ver = referent["ver"]
            if ver == 1:
                logger.debug(f"Reference {refferer!r} is v1 (implied)")
                return file, referent["path"]
            elif ver == 2:
                logger.debug(f"Reference {refferer!r} is v2")
                path = os.path.join(os.path.dirname(refferer), referent["rel"])
                path = os.path.normpath(path)
                return file, path
            raise ValueError("Unrecognized Version")

        files = tmap(_get_referent, files, Nt=1)  # self.config.concurrency)

        def _update(file, referent):
            referent = referent.strip("\n")
            refferer = file["rpath"]
            # Get the original information
            row = db.execute(
                """
                SELECT * FROM items 
                WHERE rpath = ? AND NOT isref""",
                (referent,),
            ).fetchone()

            if not row:
                txt = f"WARNING: File {refferer!r} references {referent!r} "
                txt += "but it is missing. Will just be treated as deleted"
                logger.warning(txt)
                row = DFBDST.fullrow2dict(file)
                row["size"] = -1
                return row

            row = DFBDST.fullrow2dict(row)
            row.pop("Size", None)
            # Reset some values
            row["apath"] = file["apath"]
            row["timestamp"] = file["timestamp"]
            row["isref"] = 1  # Resolved reference
            row["ref_rpath"] = refferer
            return row

        files = map(star(_update), files)

        # Insert into DB in the main thread.
        for ii, file in enumerate(files):
            db.execute(
                f"REPLACE INTO items VALUES ({','.join('?' for _ in DFBDST.COLS)})",
                DFBDST.dict2fullrow(file),
            )
            logger.info(f"updated reference {file['ref_rpath']!r}")
            # Fetching references is slow so commit more often to make sure we don't
            # lose much. Plus, we are waiting for updates
            if ii % 10 == 0:
                db.commit()
        db.commit()
        db.close()

    def _load_snapshots(self):
        # May have to rethink this for memory but that would be a *lot* of files
        self._snapshot_dict = {}
        self._snapshot_dict_ref = {}

        logger.debug("Loading snapshots from remote")

        rc = self.config.rc
        rc.start()

        # Do a sync all at once so it can be threaded with rclone directly
        params = {}
        params["_config"] = {
            "SizeOnly": True,  # These files shouldn't change. No need to worry about mtime
            "UseListR": True,  # Not too many files and could be much better
            "Transfers": self.config.concurrency,
        }

        params["srcFs"] = rcpathjoin(self.config.dst, ".dfb/snapshots")
        snap_dest = self.dbcache_dir / self.config.config_id
        params["dstFs"] = str(snap_dest)
        try:
            logger.debug(f"sync snaps with {params = }")
            rc.call("sync/sync", params=params)
        except rclonerc.RcloneError:
            logger.info("Unable to load snapshots from remote to accelerate refresh.")
            logger.debug(f"Couldn't load {params['srcFs']!r}")
            return

        for snap in sorted(snap_dest.rglob("**/*.jsonl*"), key=lambda p: p.name):
            c = 0
            for line in smart_open(str(snap)):
                line = json.loads(line)
                if (
                    line.get("_action", None) in {"prune", "comment"}
                    or line["size"] < 0
                ):
                    continue

                if line.get("isref", False):
                    self._snapshot_dict_ref[line["ref_rpath"]] = line
                else:
                    self._snapshot_dict[line["rpath"]] = line

                c += 1
            logger.info(f"Loaded {c} items from {snap.name}")

    def _apply_snapshot_file(self, file):
        if file.get("isref", False) == 2:
            if snapfile := self._snapshot_dict_ref.get(file["rpath"]):
                logger.info(
                    f"Updated reference for {file['rpath']!r} "
                    f"from {snapfile['rpath']!r}"
                )
                keep = {
                    "apath": file["apath"],
                    "timestamp": file["timestamp"],
                    "ref_rpath": file["rpath"],
                    "isref": 1,
                }

                file.clear()
                file |= snapfile | keep

            return file
        # ONLY apply if the file is listed already. Otherwise, see dbimport. This
        # includes ignore prune entries
        if not (snapfile := self._snapshot_dict.get(file["rpath"], None)):
            return file

        # They should be the same by rpath but just do a quick sanity check.
        # Include checking the timestamp for some odd files

        if (
            file["size"] != snapfile["size"]
            or file["timestamp"] != snapfile["timestamp"]
        ):
            logger.warning(
                f"snapshot for rpath = {file['rpath']!r} does not match "
                "as expected. Ignoring"
            )
            return file

        # Update and reset the file
        file |= snapfile
        file["dstinfo"] = False
        return file

    def dbimport(self, exportfiles, exportdirs, reset=False, upload=False):
        if self.config.disable_refresh:
            logger.error("DB Import not allowed due to 'disable_refresh = True")
            logger.error("Run with `--override 'disable_refresh = False'` to override")
            raise ValueError("Refresh Disabled")

        if reset:
            self.dbpath.unlink()
            self.init()

        rc = self.config.rc.start()

        # Parallelize downloading all export files. Download the specified ones and sync
        # the directories. Will then walk the file system and load them all. Note that
        # order doesn't matter for entries except prunes. So will load all entries in
        # whatever final order and then do prunes. Use random names so that there isn't
        # a conflict with the same filename

        imp = self.config.tmpdir / "import" / randstr(8)
        imp.mkdir(exist_ok=False, parents=True)

        def _dl_exportfile(exportfile):
            logger.debug(f"Downloading file {exportfile!r}")
            rc.copyfile(
                src=exportfile,
                dst=(str(imp), f"{randstr(6)}/{os.path.basename(exportfile)}"),
                _config={"NoCheckDest": True},
            )

        for _ in tmap(_dl_exportfile, exportfiles, Nt=self.config.concurrency):
            pass

        # Do these serially since parallel will be done later
        for exportdir in exportdirs:
            logger.debug(f"Downloading directory {exportdir!r}")
            params = {}
            params["_config"] = {
                "IgnoreTimes": True,  # Always download
                "UseListR": True,  # Not too many files and could be much better
                "Transfers": self.config.concurrency,
            }

            params["srcFs"] = exportdir
            params["dstFs"] = rcpathjoin(str(imp), randstr(6))
            rc.call("sync/sync", params=params)

        # The "sorted" is not really needed but it is likely to keep the
        # database cleaner
        loadfiles = sorted(imp.rglob("*.jsonl*"), key=lambda file: file.name)

        prune = []  # prune is OUTSIDE the file loop as noted above
        for exportfile in loadfiles:
            logger.debug(f"Importing from {str(exportfile)!r}")
            files = []
            pcount = 0
            with smart_open(exportfile, "rt") as fp:
                for line in fp:
                    file = json.loads(line)
                    if file.get("_action", None) == "comment":
                        continue

                    if file.get("_action", None) == "prune":
                        prune.append(file)
                        pcount += 1
                        continue

                    files.append(file)

            with self.db() as db:
                db.executemany(
                    f"""INSERT OR REPLACE INTO items 
                        VALUES ({','.join('?' for _ in DFBDST.COLS)})""",
                    [DFBDST.dict2fullrow(file) for file in files],
                )
            msg = f"  Imported {len(files)} files"
            if pcount:
                msg += f" and will prune {pcount}"
            logger.info(msg)

        if prune:
            with self.db() as db:
                db.executemany(
                    """
                    DELETE FROM items 
                    WHERE rpath = ?""",
                    ((file["rpath"],) for file in prune),
                )
            logger.info(f"Pruned {len(prune)} files from all exports")

        if upload:
            snapdir = self.snap_file.parent / self.snap_file.stem
            snapdir.mkdir(parents=True)  # Shouldn't already exists

            for ii, loadfile in enumerate(loadfiles):
                dst = snapdir / f"{ii}.{loadfile.name}"
                shutil.copy2(loadfile, dst)

            self.push_snapshots(compress=False)

    def insert_or_replace_many(self, files, *, insert, replace):
        """
        Allows inserting or replacing. This requires being explicit to avoid wrong
        insertions
        """
        action = []
        if insert:
            action.append("INSERT")
        if replace:
            action.append("REPLACE")
        action = " OR ".join(action)
        sql = f"{action} INTO items VALUES ({','.join('?' for _ in DFBDST.COLS)})"

        # Collect them all. We will do it anyway in the DB and this way it can be yielded
        files = list(files)
        # Insert into DB in the main thread
        rows = map(DFBDST.dict2fullrow, files)
        # ALWAYS wait before an executemany since that could lock the DB
        rows = list(rows)

        db = self.db()
        with db:
            db.executemany(sql, rows)
        # db.commit()
        # db.close()

        with self.snap_file.open(mode="at") as fp:
            for file in files:
                print(json.dumps(file), file=fp, flush=True)

        return files

    insert_many = partialmethod(insert_or_replace_many, insert=True, replace=False)
    replace_many = partialmethod(insert_or_replace_many, insert=False, replace=True)

    def _insert_or_replace(self, file, *, insert, replace):
        """
        Allows inserting or replacing. This requires being explicit to avoid wrong
        insertions.

        These are also partialmethods defined below

            Insert only: db._insert_or_replace(file,insert=True,replace=False)
            Insert only: db._insert_or_replace(file,insert=False,replace=True)
            both:      : db._insert_or_replace(file,insert=True,replace=True)

        """
        return self.insert_or_replace_many([file], insert=insert, replace=replace)

    insert = partialmethod(_insert_or_replace, insert=True, replace=False)
    replace = partialmethod(_insert_or_replace, insert=False, replace=True)
    insert_or_replace = partialmethod(_insert_or_replace, insert=True, replace=True)

    def _snapshot_query_builder(
        self,
        *,
        path="",
        before=None,
        after=None,
        select="*",
        groupselect="*",
        export=False,
        remove_delete=True,
        delete_only=False,
        conditions=None,
        query_prefix="snap",
    ):
        """
        Build a query for snapshots. This can then be evaluated later directory or as a
        subquery.

        Inputs:
        -------

        path: ''
            Starting path

        before:
            Select files <= before. This is the "at" snapshot time. Will be parsed by
            timestamp_parser. Times are inclusive on both ends

        after:
            Select files >= after. This is the "at" snapshot time. Will be parsed by
            timestamp_parser. Times are inclusive on both ends

        select
            What to return. Used after the GROUP BY statement

        groupselect
            Select inside the GROUP BY statement. Useful for additional values
            but MUST also include "*"

        export [False]
            If True, includes multiples. Will override remove_delete and delete_only

        remove_delete: [True]
            If False, will keep deleted items. Uses a subquery which should be faster
            than manual filtering. If used with delete_only, will get nothing.

        delete_only [False]
            Only show deleted items. If used with remove_delete, will get nothing

        conditions:
            List of additional (sql,param_dict) pairs. Warning: Do not let sql be user input.
            Examples: ('apparentparent LIKE :path',{'path':'a/sub/path/'})

            WARNING: Do not do ('size >= ?',0) since that will then include the non-deleted
                     version. It is better to filter it later.

        query_prefix: Prefix to be used for all query parameters. Can be useful if building
                 subqueries

        Returns:
        --------
        query : Text of the query
        params: Dictionary of the parameter substitutions

        """
        qp = query_prefix

        conditions = conditions or []
        if "*" not in groupselect:
            raise ValueError("groupselect must have '*'")

        query = []
        params = {}

        if export:
            remove_delete = delete_only = False

        # Build the snapshot. Note that the select is never *user*
        # specified so there isn't an SQL injection risk.

        # Always build as a nested query even if not needed. The query optimizer
        # should flatten it. This makes the dynamic construction much easier.
        # From ChatGPT:
        #    While the performance impact of using SELECT * FROM (SELECT * FROM items)
        #    is generally negligible in SQLite due to query optimization, it's a good
        #    practice to avoid such constructs unless they serve a specific purpose,
        #    such as when dynamically building complex queries.

        query.extend(
            [
                "SELECT",
                groupselect,
                "FROM items",
            ]
        )

        if path:
            path = path.removesuffix("/").removeprefix("./")
            conditions.append((f"apath LIKE :{qp}_path", {f"{qp}_path": f"{path}/%"}))

        if before:
            b0 = before
            before = timestamp_parser(
                before,
                aware=True,
                epoch=True,
                now=self.config.now.obj,
            )
            logger.debug(f"Interpreted before = {b0} as {before} (s)")
            conditions.append((f"timestamp <= :{qp}_before", {f"{qp}_before": before}))

        if after:
            a0 = after
            after = timestamp_parser(
                after,
                aware=True,
                epoch=True,
                now=self.config.now.obj,
            )
            logger.debug(f"Interpreted after = {a0} as {after} (s)")
            conditions.append((f"timestamp >= :{qp}_after", {f"{qp}_after": after}))

        if conditions:
            query.append("WHERE")
            query.append(" AND ".join(cond[0] for cond in conditions))
            params |= {k: v for c in conditions for k, v in c[1].items()}

        if not export:
            query.append("GROUP BY apath HAVING MAX(timestamp)")
        # query.append("ORDER BY LOWER(apath)")

        query = "\n".join(query)

        outq_cond = []
        if remove_delete:
            outq_cond.append("size >= 0")
        if delete_only:
            outq_cond.append("size < 0")

        # Build the nested query even if not needed to capture the different
        # select and groupselect
        query = "-- Subquery with everything that is then filtered\n" + query
        query = f"SELECT {select} FROM (\n{indent(query,' '*4)}\n)"
        if outq_cond:
            query += "\nWHERE\n" + " AND ".join(outq_cond)

        return query, params

    def snapshot(self, add_query="", **kwargs):
        """
        See _snapshot_query_builder for help. Can specify additional queries but must be
        in select.
        """
        query, params = self._snapshot_query_builder(**kwargs)
        if add_query:
            query += "\n" + add_query

        db = self.db()
        with db:
            r = db.execute(query, params)
        return r

    def ls(
        self,
        subdir="",
        *,
        before=None,
        after=None,
        select="*",
        remove_delete=True,
        delete_only=False,
        conditions=None,
        recursive=False,
    ):
        """

        Build a query.

        path: ''
            Starting path

        before:
            Select files <= before. This is the "at" snapshot time. Will be parsed by
            timestamp_parser. Times are inclusive on both ends

        after:
            Select files >= after. This is the "at" snapshot time. Will be parsed by
            timestamp_parser. Times are inclusive on both ends

        select
            What to return.

        remove_delete: [True]
            If False, will keep deleted items. Uses a subquery which should be faster
            than manual filtering. If used with delete_only, will get nothing.

        delete_only [False]
            Only show deleted items. If used with remove_delete, will get nothing

        conditions:
            List of additional (sql,val) pairs. Warning: Do not let sql be user input.
            Examples: ('apparentparent LIKE ?','a/sub/path/')

            WARNING: Do not do ('size >= ?',0) since that will then include the non-deleted
                     version. It is better to filter it later.

        recursive: [False]
            List all items, not just the one directory

        Some of this very clever SQL came from my reddit post here:
        https://www.reddit.com/r/sqlite/comments/123bivr/comment/jdu9xvl/?context=3
        """
        subdir = subdir.removeprefix("./").removesuffix("/")
        db = self.db()

        # This method makes essentially two queries that look a lot like snapshots
        #     1. Snapshot like query that then filters using the above-noted clever SQL
        #        to filter subdirs. (The old version of this did something similar w/o
        #        filters for the latest then checked each dir. This is more efficient)
        #
        #     2. Files: Add conditions to apath to make sure it only lists in that
        #        parent directory

        conditions = conditions or []

        ## Files
        fcond = conditions.copy()

        if not recursive:
            fcond.append(
                [
                    "apath NOT LIKE :onedepth",
                    {"onedepth": os.path.join(subdir, "%", "%")},
                ]
            )

        groupselect = dedent(
            """
            *, 
            COUNT(*) AS versions,
            SUM(
                CASE  
                    WHEN size > 0 THEN size 
                    ELSE 0
                END
            ) as tot_size -- Need to account for -1 vals
            """
        )

        fquery, fparams = self._snapshot_query_builder(
            path=subdir,
            before=before,
            after=after,
            select="*",
            groupselect=groupselect,
            remove_delete=remove_delete,
            delete_only=delete_only,
            conditions=fcond,
        )

        files = [DFBDST.fullrow2dict(r) for r in db.execute(fquery, fparams)]

        ## Directories.
        if recursive:
            # Do this in Python as it is cleaner than SQL
            directories = {os.path.dirname(file["apath"]) for file in files}

            for directory in directories.copy():
                if not directory:  # At root
                    continue

                directory = os.path.relpath(directory, subdir)  # remove subdir
                while directory:
                    if directory := os.path.dirname(directory):
                        directories.add(os.path.join(subdir, directory))

            directories.difference_update({"", "./"})
        else:
            # Use the snapshot query builder with all of the conditions to make a query with
            # all valid files. Then use the fancy SQL to down-select directories. If it is a
            # subdir, it needs an additional filter to remove the subdir from the apath names.
            # The "QQQQ" sub is purely cosmetic to get the indents of the subquery *after* the
            # dedents of the outer query
            dir_query, dir_params = self._snapshot_query_builder(
                path=subdir,
                before=before,
                after=after,
                select="apath",
                remove_delete=remove_delete,
                delete_only=delete_only,
                conditions=conditions,
            )

            params = dir_params.copy()
            if subdir:
                query = dedent(
                    f"""
                    WITH
                        snappaths AS (
                        QQQQ
                        ),
                        subpaths AS (
                            SELECT SUBSTR(snappaths.apath, {len(subdir) + 2}) AS apath
                            FROM snappaths
                            WHERE snappaths.apath LIKE :subdir
                        )
                        
                    """
                ).replace("QQQQ", indent(dir_query, " " * 8))
                params["subdir"] = f"{subdir}/%"
            else:
                query = dedent(
                    f"""
                    WITH
                        subpaths AS (
                        QQQQ
                        )"""
                ).replace("QQQQ", indent(dir_query, " " * 8))

            query += dedent(
                """
                -- Get just the next path element
                -- https://www.reddit.com/r/sqlite/comments/123bivr/comment/jdu9xvl/?context=3
                SELECT DISTINCT 
                        SUBSTR(
                            apath,
                            1,
                            CASE INSTR(apath, '/')
                                WHEN 0
                                THEN LENGTH(apath)
                                ELSE INSTR(apath, '/')
                            END
                        ) AS sub
                FROM subpaths
                """
            )
            apaths = (r["sub"] for r in db.execute(query, params))
            apaths = (apath for apath in apaths if apath.endswith("/"))
            directories = [os.path.join(subdir, apath) for apath in apaths]

        return directories, files

    def file_versions(self, filepath, count_refs=False):
        db = self.db()
        with db:
            versions = db.execute(
                "SELECT * FROM items WHERE apath = ? ORDER BY timestamp", (filepath,)
            )
        versions = [self.fullrow2dict(v) for v in versions]

        if count_refs:
            for file in versions:
                counts = db.execute(
                    """
                    SELECT COUNT(rpath) AS count FROM items
                    WHERE rpath = ?""",
                    (file["rpath"],),
                ).fetchone()
                file["ref_count"] = counts.get("count", default=0)

        db.close()
        return versions

    def group_by_apath(self, select="*", conditions=None):
        """
        Group by apath where each group will be sorted by timestamp.
        (so you can use bisect to quickly find elements)

        Inputs:
        ------

        select
            What to return. MUST include 'apath'

        conditions:
            NOTE: different than in ls
            query_txt, qvals such as
                query_txt = "WHERE size > ? and size < ?"
                qvals = [100,200]


        """
        db = self.db()

        if conditions:
            qtxt, qvals = conditions
            qvals = listify(qvals)
        else:
            qtxt, qvals = "", []

        with db:
            Qres = db.execute(
                f"""
                SELECT {select} FROM items
                {qtxt}
                ORDER BY
                    LOWER(apath),timestamp""",
                qvals,
            )
            Qres = map(DFBDST.fullrow2dict, Qres)

        row = next(Qres)
        try:
            name = row["apath"]
        except KeyError:
            raise ValueError("Must include 'apath' in 'select'")
        group = [row]

        for row in Qres:
            if row["apath"] == name:
                group.append(row)
            else:
                yield name, group
                group = [row]
                name = row["apath"]
        yield name, group  # Last item

    def push_snapshots(self, compress=True):
        """
        Compress (optional) and push all snapshots.
        This will catch ones from interrupted runs
        """
        for snap_file in self.config.snap_cache_dir.rglob("**/*.jsonl"):
            if snap_file.stat().st_size == 0:
                logger.debug(f"Empty snapshot {str(snap_file)!r}. Unlink")
                snap_file.unlink()
            elif compress:
                snapz = snap_file.with_suffix(".jsonl.gz")
                with gz.open(str(snapz), "wb") as fz, snap_file.open("rb") as fu:
                    shutil.copyfileobj(fu, fz)
                snap_file.unlink()
                logger.debug(f"Compressed {str(snap_file)!r} to {str(snapz)!r}")

        rc = self.config.rc
        rc.start()

        params = {
            "createEmptySrcDirs": False,
            "deleteEmptySrcDirs": True,  # Cleans local
        }
        params["_config"] = {
            "NoCheckDest": True,  # We know they shouldn't exists
            "NoTraverse": True,
            "NoUpdateDirModTime": True,  # Don't care
            "Transfers": self.config.concurrency,
            "Metadata": self.config.metadata,
        }
        params["srcFs"] = str(self.config.snap_cache_dir)
        params["dstFs"] = rcpathjoin(self.config.dst, ".dfb/snapshots")

        logger.debug(f"moving files from {params['srcFs']!r} to {params['dstFs']!r}")
        rc.call("sync/move", params=params)

    @classmethod
    def dict2fullrow(cls, rowdict):
        """Take a dict of the file and convert it to a DB row"""
        rowdict = rowdict.copy()

        if cs := rowdict.get("checksum", None):
            rowdict["checksum"] = json.dumps(cs)

        row = [rowdict.pop(key, None) for key, _ in cls.COLS[:-1]]
        row.append(json.dumps(rowdict) if rowdict else None)  # remain
        return row

    @staticmethod
    def fullrow2dict(row):
        row = dict(row)

        try:
            row["checksum"] = json.loads(row["checksum"])
        except (KeyError, TypeError, json.JSONDecodeError):
            pass

        if remain := row.pop("remain", None):
            row.update(json.loads(remain))

        return row



import os
import sys
import json

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

from dfb.cli import cli

# Local
import testutils
from testutils import Capture

# testing
import pytest


def test_apath2rpath_cli():
    files = ["name1.ext", "path/to/name2.txt.gz"]
    with Capture() as cap:
        cli(["utils", "apath2rpath", *files, "--date", "20120304T05:06:07Z"])
    assert (
        cap.out.strip()
        == "name1.20120304050607.ext\npath/to/name2.20120304050607.txt.gz"
    )

    with Capture() as cap:
        cli(["utils", "apath2rpath", *files, "--date", "20120304T05:06:07Z", "-0"])
    assert (
        cap.out_bytes.strip()
        == b"name1.20120304050607.ext\x00path/to/name2.20120304050607.txt.gz"
    )


def test_rpath2apath_cli():
    file = "name1.20180811094630.ext"
    with Capture() as cap:
        cli(["utils", "rpath2apath", "name1.20180811094630.ext"])
    assert {
        "apath": "name1.ext",
        "timestamp": "2018-08-11T09:46:30+00:00",
        "flag": "",
    } == json.loads(cap.out)

    with Capture() as cap:
        cli(["utils", "rpath2apath", "name1.20180811094630D.ext"])
    assert {
        "apath": "name1.ext",
        "timestamp": "2018-08-11T09:46:30+00:00",
        "flag": "D",
    } == json.loads(cap.out)

    with Capture() as cap:
        cli(["utils", "rpath2apath", "name1.20180811094630R.ext"])
    assert {
        "apath": "name1.ext",
        "timestamp": "2018-08-11T09:46:30+00:00",
        "flag": "R",
    } == json.loads(cap.out)

    with Capture() as cap:
        cli(
            [
                "utils",
                "rpath2apath",
                "first.20180811094630.ext",
                "second.20220625232247",
            ]
        )

    assert [
        {"apath": "first.ext", "timestamp": "2018-08-11T09:46:30+00:00", "flag": ""},
        {"apath": "second", "timestamp": "2022-06-25T23:22:47+00:00", "flag": ""},
    ] == [json.loads(line) for line in cap.out.splitlines()]


if __name__ == "__main__":
    test_apath2rpath_cli()
    test_rpath2apath_cli()

    print("=" * 50)
    print(" All Passed ".center(50, "="))
    print("=" * 50)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys
from pathlib import Path
import gzip as gz
import subprocess
import json
import shlex
import shutil
from textwrap import dedent

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

# Local
import testutils
from testutils import Capture

# testing
import pytest


def test_listing():
    test = testutils.Tester(name="listing")

    test.config["renames"] = "mtime"
    test.write_config()

    test.write_pre("src/untouched.txt", "Never Modified")
    test.write_pre("src/sub1/file.txt", "Always move")
    test.write_pre("src/new1.txt", "new 1")
    test.write_pre("src/mod.txt", ".")
    test.write_pre("src/n1d3.txt", "new at 1, delete at 3")
    test.backup("-q", offset=1)

    test.move("src/sub1/file.txt", "src/sub2/file.txt")
    test.write_pre("src/new2.txt", "new 2")
    test.write_pre("src/mod.txt", "..")
    test.backup("-q", offset=2)

    test.move("src/sub2/file.txt", "src/sub3/file.txt")
    test.write_pre("src/new3.txt", "new 3")
    test.write_pre("src/mod.txt", "...")
    os.unlink("src/n1d3.txt")
    test.backup("-q", offset=3)

    test.move("src/sub3/file.txt", "src/sub4/file.txt")
    test.write_pre("src/new4.txt", "new 4")
    test.write_pre("src/mod.txt", "....")
    test.backup("-q", offset=4)

    # # # ls
    out = test.ls("--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "new4.txt",
        "sub4/",
        "untouched.txt",
    }

    out = test.ls("--at", "u3.1", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "sub3/",
        "untouched.txt",
    }

    out = test.ls("--at", "u2.1", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "new1.txt",
        "n1d3.txt",
        "new2.txt",
        "sub2/",
        "untouched.txt",
    }

    out = test.ls("--at", "u1.1", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"mod.txt", "new1.txt", "n1d3.txt", "sub1/", "untouched.txt"}

    ## rpath
    out = test.ls("--no-header", "--real-path")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.19700101000004.txt",
        "new1.19700101000001.txt",
        "new2.19700101000002.txt",
        "new3.19700101000003.txt",
        "new4.19700101000004.txt",
        "sub4/",
        "untouched.19700101000001.txt",
    }

    out = test.ls("--at", "u1.1", "--no-header", "--rpath")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.19700101000001.txt",
        "n1d3.19700101000001.txt",
        "new1.19700101000001.txt",
        "sub1/",
        "untouched.19700101000001.txt",
    }

    # include header to check it says 'rpath'
    out = test.ls("--at", "u1.1", "sub1", "--real-path", "--full-path")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"rpath", "sub1/file.19700101000001.txt"}

    # references. Should point to the ref_rpath, not the regular rpath
    assert (
        not "../sub1/file.19700101000001.txt"
        == test.ls("sub2", "--before", "u2.5", "--rpath", "--no-header").strip()
    )
    assert (
        not "sub1/file.19700101000001.txt"
        == test.ls(
            "sub2", "--before", "u2.5", "--rpath", "--no-header", "--full-path"
        ).strip()
    )

    assert test.ls("sub4", "--no-header").strip() == "file.txt"
    assert (
        test.ls("sub4", "--no-header", "--real-path").strip()
        == "file.19700101000004R.txt"
    )
    assert (
        test.ls("sub4", "--no-header", "--rpath", "--rpath").strip()
        == "../sub1/file.19700101000001.txt"
    )
    assert (
        test.ls("sub4", "--no-header", "--rpath", "--rpath", "--full-path").strip()
        == "sub1/file.19700101000001.txt"
    )

    ## Recursive and only

    out = test.ls("-d", "--no-header", "--recursive", "--list", "both")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "n1d3.txt (DEL)",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "new4.txt",
        "sub1/",
        "sub1/file.txt (DEL)",
        "sub2/",
        "sub2/file.txt (DEL)",
        "sub3/",
        "sub3/file.txt (DEL)",
        "sub4/",
        "sub4/file.txt",
        "untouched.txt",
    }

    out = test.ls("-d", "--no-header", "--recursive")  # default to --list files
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "n1d3.txt (DEL)",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "new4.txt",
        "sub1/file.txt (DEL)",
        "sub2/file.txt (DEL)",
        "sub3/file.txt (DEL)",
        "sub4/file.txt",
        "untouched.txt",
    }

    out = test.ls("-d", "--no-header", "--recursive", "--list-only", "files")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "n1d3.txt (DEL)",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "new4.txt",
        "sub1/file.txt (DEL)",
        "sub2/file.txt (DEL)",
        "sub3/file.txt (DEL)",
        "sub4/file.txt",
        "untouched.txt",
    }

    out = test.ls("-d", "--no-header", "--list-only", "dirs")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "sub1/",
        "sub2/",
        "sub3/",
        "sub4/",
    }

    # with rpath
    out = test.ls("--del", "-r", "--rpath", "--list-only", "files", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "mod.19700101000004.txt",
        "n1d3.19700101000003D.txt (DEL)",
        "new1.19700101000001.txt",
        "new2.19700101000002.txt",
        "new3.19700101000003.txt",
        "new4.19700101000004.txt",
        "sub1/file.19700101000002D.txt (DEL)",
        "sub2/file.19700101000003D.txt (DEL)",
        "sub3/file.19700101000004D.txt (DEL)",
        "sub4/file.19700101000004R.txt",
        "untouched.19700101000001.txt",
    }

    ## Head and tails

    out = test.ls("-d", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    allitems = {
        "mod.txt",
        "n1d3.txt (DEL)",
        "new1.txt",
        "new2.txt",
        "new3.txt",
        "new4.txt",
        "sub1/",
        "sub2/",
        "sub3/",
        "sub4/",
        "untouched.txt",
    }
    assert items == allitems

    # Test head and tail. Note that these are sorted so this works file
    out = test.ls("-d", "--head", "2")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "mod.txt", "n1d3.txt (DEL)"}

    out = test.ls("-d", "--head", "2", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "mod.txt", "n1d3.txt (DEL)"} - {"path"}

    out = test.ls("-d", "--tail", "2")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "sub4/", "untouched.txt"}

    out = test.ls("-d", "--tail", "2", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "sub4/", "untouched.txt"} - {"path"}

    out = test.ls("-d", "--head", "1", "--tail", "2")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "untouched.txt", "sub4/", "...", "mod.txt"}

    out = test.ls("-d", "--head", "1", "--tail", "2", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"path", "untouched.txt", "sub4/", "...", "mod.txt"} - {"path"}

    out = test.ls("-d", "--head", "9", "--tail", "2", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == allitems

    out = test.ls("-d", "--head", "12", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == allitems

    out = test.ls("-d", "--head", "99", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == allitems

    out = test.ls("-d", "--tail", "12", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == allitems

    out = test.ls("-d", "--tail", "99", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == allitems

    out1 = test.ls("-d", "--at", "u2.1")
    out2 = test.ls("-d", "--at", "u2.1", "--no-header")
    items = {i.strip() for i in out2.split("\n") if i.strip()}
    assert items == {
        "mod.txt",
        "n1d3.txt",
        "new1.txt",
        "new2.txt",
        "sub1/",
        "sub2/",
        "untouched.txt",
    }

    # Before and afters and only. Also test using a delta
    out = test.ls("--no-header", "--before", "u3", "--after", "u3")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"mod.txt", "new3.txt", "sub3/"}

    out = test.ls("--no-header", "--only", "u3")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"mod.txt", "new3.txt", "sub3/"}

    # Test using a delta. Use offset so this is u3
    out = test.ls("--only", "3.5 seconds", "--no-header", offset=6.5)
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"mod.txt", "new3.txt", "sub3/"}

    out = test.ls("sub1", "-d", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"file.txt (DEL)"}

    out = test.ls("sub1", "-d", "--full-path", "--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"sub1/file.txt (DEL)"}

    # I am going to stop checking output on all since the modtimes will mess it up but this
    # will test the code path and I have verified it manually
    out = test.ls("-l")
    out = test.ls("-ll")
    out = test.ls("-lld")
    out = test.ls("-lld", "--timestamp-local", "--human")

    out = test.ls("sub1dasdas")

    # Timestamps

    with Capture() as cap:
        test.call("timestamps")
    assert [l.strip() for l in cap.out.splitlines()] == [
        l.strip()
        for l in """\
                   Timestamp  Total  Deleted  Moved  Size
        1970-01-01T00:00:01Z      5        0      0  52
        1970-01-01T00:00:02Z      4        1      1  7
        1970-01-01T00:00:03Z      5        2      2  8
        1970-01-01T00:00:04Z      4        1      2  9
        """.strip().splitlines()
    ]

    with Capture() as cap:
        test.call("timestamps", "sub4")
    assert [l.strip() for l in cap.out.splitlines()] == [
        l.strip()
        for l in """\
                   Timestamp  Total  Deleted  Moved  Size
        1970-01-01T00:00:04Z      1        0      1  0
        """.strip().splitlines()
    ]

    with Capture() as cap:
        test.call("timestamps", "sub2")
    assert [l.strip() for l in cap.out.splitlines()] == [
        l.strip()
        for l in """\
                     Timestamp  Total  Deleted  Moved  Size
          1970-01-01T00:00:02Z      1        0      1  0
          1970-01-01T00:00:03Z      1        1      1  0
        """.strip().splitlines()
    ]

    test.call("timestamps", "--human")

    test.call("timestamps")
    test.call("timestamps", "--human", "--timestamp-local")

    # Verify head and tail. Test just the code paths for now.
    with Capture() as cap:
        test.call("timestamps", "--head", "1")
    assert len(cap.out.splitlines()) == 2

    with Capture() as cap:
        test.call("timestamps", "--tail", "1")
    assert len(cap.out.splitlines()) == 2

    with Capture() as cap:
        test.call("timestamps", "--head", "1", "--tail", "1")
    assert len(cap.out.splitlines()) == 4
    assert "..." in cap.out

    with Capture() as cap:
        test.call("timestamps", "--head", "1000", "--tail", "1")
    assert len(cap.out.splitlines()) == 5
    assert "..." not in cap.out

    test.call("timestamps", "--tail", "1")
    test.call("timestamps", "--head", "1", "--tail", "1")

    # +=
    with Capture() as cap:
        test.call("snapshot")
    clisnap = [json.loads(line) for line in cap.out.splitlines() if line]
    clisnap = {f["apath"] for f in clisnap}

    assert clisnap == {dict(f)["apath"] for f in test.remote_snapshot()}
    # -

    for ts in [1, 2, 3, 4]:
        test.call("snapshot", "--at", f"u{ts+0.1}", "--output", "tmp.jsonl")
        clisnap = [json.loads(line) for line in open("tmp.jsonl") if line]
        clisnap = {f["apath"] for f in clisnap}
        assert clisnap == {
            dict(f)["apath"] for f in test.remote_snapshot(before=ts + 0.1)
        }

    # Like ls -l, this is hard to verify with ModTime. Just testing code paths.
    # It has been verified that it works
    test.call("versions", "mod.txt")
    test.call("versions", "mod.txt", "--timestamp-local")
    test.call("versions", "mod.txt", "--ref-count")

    test.call("versions", "sub1/file.txt", "--ref-count", "--real-path")
    test.call("versions", "sub2/file.txt", "--ref-count", "--real-path")
    test.call("versions", "sub3/file.txt", "--ref-count", "--real-path")
    test.call("versions", "sub4/file.txt", "--ref-count", "--real-path")
    test.call("versions", "sub4/file.txt", "--ref-count", "--real-path", "--real-path")

    test.call("versions", "made up file", "--ref-count", "--real-path")

    # This is also hard to test so just use the number of lines
    with Capture() as cap:
        test.call("versions", "mod.txt")
    assert len(cap.out.splitlines()) == 6

    with Capture() as cap:
        test.call("versions", "mod.txt", "--no-header")
    assert len(cap.out.splitlines()) == 5

    with Capture() as cap:
        test.call("versions", "mod.txt", "--head", "1")
    assert len(cap.out.splitlines()) == 3

    with Capture() as cap:
        test.call("versions", "mod.txt", "--no-header", "--head", "1")
    assert len(cap.out.splitlines()) == 2

    with Capture() as cap:
        test.call("versions", "mod.txt", "--tail", "1")
    assert len(cap.out.splitlines()) == 3

    with Capture() as cap:
        test.call("versions", "mod.txt", "--no-header", "--tail", "1")
    assert len(cap.out.splitlines()) == 2

    with Capture() as cap:
        test.call("versions", "mod.txt", "--head", "1", "--tail", "1")
    assert len(cap.out.splitlines()) == 5
    assert "..." in cap.out

    with Capture() as cap:
        test.call("versions", "mod.txt", "--no-header", "--head", "1", "--tail", "1")
    assert len(cap.out.splitlines()) == 4
    assert "..." in cap.out

    with Capture() as cap:
        test.call("versions", "mod.txt", "--head", "99", "--tail", "1")
    assert len(cap.out.splitlines()) == 6
    assert "..." not in cap.out

    # timestamp-include-filters
    cmd = ["advanced", "timestamp-include-filters"]
    cmd += ["--after", "u2", "--before", "u3"]
    with Capture() as cap:
        test.call(*cmd)
    assert [
        "--include",
        "*.19700101000002*",
        "--include",
        "*.19700101000003*",
    ] == shlex.split(cap.out)

    cmd = ["advanced", "timestamp-include-filters"]
    cmd += ["--after", "u2", "--before", "u3", "sub3"]  # add subdir to test
    with Capture() as cap:
        test.call(*cmd)
    assert ["--include", "*.19700101000003*"] == shlex.split(cap.out)

    ## Summary
    with Capture() as cap:
        test.call("summary")
    assert [l.strip() for l in cap.out.splitlines()] == [
        l.strip()
        for l in """\
                Path:  '/'
               After:  <<earliest>>
              Before:  <<latest>>
          Timestamps:  4
               Total:  18
             Deleted:  4
               Moved:  5
                Size:  76 (76 B)
        """.strip().splitlines()
    ]

    with Capture() as cap:
        test.call("summary", "--after", "u2")
    assert [l.strip() for l in cap.out.splitlines()] == [
        l.strip()
        for l in """\
                Path:  '/'
               After:  'u2'
              Before:  <<latest>>
          Timestamps:  3
               Total:  13
             Deleted:  4
               Moved:  5
                Size:  24 (24 B)
        """.strip().splitlines()
    ]


def test_del():
    test = testutils.Tester(name="listing_del")

    test.config["renames"] = "mtime"
    test.write_config()

    test.write_pre("src/untouched.txt", "Never Modified")
    test.write_pre("src/del_at_3.txt", "delete at 3")
    test.write_pre("src/sub_del_at_3/file.txt", "delete at 3 --sub")
    test.write_pre("src/sub_del_at_5/file.txt", "delete at 5 --sub")
    test.write_pre("src/mv1/f1.txt", "move each time")
    test.write_pre("src/new1/new1.txt", "new 1")

    test.backup(offset=1)

    os.unlink("src/del_at_3.txt")
    shutil.rmtree("src/sub_del_at_3")
    test.move("src/mv1/f1.txt", "src/mv2/f2.txt")
    os.rmdir("src/mv1")
    test.write_pre("src/new3/new3.txt", "new 2")

    test.backup(offset=3)

    shutil.rmtree("src/sub_del_at_5")
    test.move("src/mv2/f2.txt", "src/mv3/f3.txt")
    os.rmdir("src/mv2")
    test.write_pre("src/new5/new5.txt", "new 2")

    test.backup(offset=5)

    out = test.ls("--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {"mv3/", "new1/", "new3/", "new5/", "untouched.txt"}

    out = test.ls("--no-header", "--del")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "del_at_3.txt (DEL)",
        "mv1/",
        "mv2/",
        "mv3/",
        "new1/",
        "new3/",
        "new5/",
        "sub_del_at_3/",
        "sub_del_at_5/",
        "untouched.txt",
    }

    out = test.ls("--no-header", "--del", "--del")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "del_at_3.txt (DEL)",
        "mv1/",
        "mv2/",
        "sub_del_at_3/",
        "sub_del_at_5/",
    }

    test.call("snapshot", "--output", "0.jsonl")
    test.call("snapshot", "--del", "--output", "1.jsonl")
    test.call("snapshot", "--del", "--del", "--output", "2.jsonl")

    with open("0.jsonl") as fp:
        i0 = {(item["apath"], item["size"]) for item in map(json.loads, fp)}
    assert i0 == {
        ("mv3/f3.txt", 14),
        ("new1/new1.txt", 5),
        ("new3/new3.txt", 5),
        ("new5/new5.txt", 5),
        ("untouched.txt", 14),
    }

    with open("1.jsonl") as fp:
        i1 = {(item["apath"], item["size"]) for item in map(json.loads, fp)}
    assert i1 == {
        ("del_at_3.txt", -1),
        ("mv1/f1.txt", -1),
        ("mv2/f2.txt", -1),
        ("mv3/f3.txt", 14),
        ("new1/new1.txt", 5),
        ("new3/new3.txt", 5),
        ("new5/new5.txt", 5),
        ("sub_del_at_3/file.txt", -1),
        ("sub_del_at_5/file.txt", -1),
        ("untouched.txt", 14),
    }

    with open("2.jsonl") as fp:
        i2 = {(item["apath"], item["size"]) for item in map(json.loads, fp)}
    assert i2 == {
        ("del_at_3.txt", -1),
        ("mv1/f1.txt", -1),
        ("mv2/f2.txt", -1),
        ("sub_del_at_3/file.txt", -1),
        ("sub_del_at_5/file.txt", -1),
    }

    assert i0.union(i2) == i1


def test_tree():
    test = testutils.Tester(name="tree")
    test.write_config()

    # Multiple on a level
    test.write_pre("src/file1.txt", "1")
    test.write_pre("src/file2.txt", "2")
    test.write_pre("src/sub1/file3.txt", "3")
    test.write_pre("src/sub1/file4.txt", "4")
    test.write_pre("src/sub1/ssub1/file5.txt", "5")
    test.write_pre("src/sub1/ssub1/file6.txt", "6")

    # single on a level
    test.write_pre("src/sub2/file7.txt", "7")
    test.write_pre("src/sub2/ssub2/file8.txt", "8")

    # skip a level
    test.write_pre("src/sub3/ssub3/file9.txt", "9")

    test.backup(offset=1)

    assert (
        test.tree().strip()
        == dedent(
            """
        /
        ├── file1.txt
        ├── file2.txt
        ├── sub1/
        │   ├── file3.txt
        │   ├── file4.txt
        │   └── ssub1/
        │       ├── file5.txt
        │       └── file6.txt
        ├── sub2/
        │   ├── file7.txt
        │   └── ssub2/
        │       └── file8.txt
        └── sub3/
            └── ssub3/
                └── file9.txt
        """
        ).strip()
    )

    assert (
        test.tree("--max-depth", "1").strip()
        == dedent(
            """
        /
        ├── file1.txt
        ├── file2.txt
        ├── sub1/
        ├── sub2/
        └── sub3/
        """
        ).strip()
    )
    assert (
        test.tree("--max-depth", "2").strip()
        == dedent(
            """
        /
        ├── file1.txt
        ├── file2.txt
        ├── sub1/
        │   ├── file3.txt
        │   ├── file4.txt
        │   └── ssub1/
        ├── sub2/
        │   ├── file7.txt
        │   └── ssub2/
        └── sub3/
            └── ssub3/
        """
        ).strip()
    )
    assert (
        test.tree("sub1").strip()
        == dedent(
            """
        sub1/
        ├── file3.txt
        ├── file4.txt
        └── ssub1/
            ├── file5.txt
            └── file6.txt
        """
        ).strip()
    )

    assert (
        test.tree("sub1", "--max-depth", "1").strip()
        == dedent(
            """
            sub1/
            ├── file3.txt
            ├── file4.txt
            └── ssub1/
            """
        ).strip()
    )

    os.unlink("src/sub1/ssub1/file5.txt")
    test.write_post("src/sub3/ssub3/file9.txt", "nine")

    test.backup(offset=3)

    assert (
        test.tree("--del").strip()
        == dedent(
            """
            /
            ├── file1.txt
            ├── file2.txt
            ├── sub1/
            │   ├── file3.txt
            │   ├── file4.txt
            │   └── ssub1/
            │       ├── file5.txt (DEL)
            │       └── file6.txt
            ├── sub2/
            │   ├── file7.txt
            │   └── ssub2/
            │       └── file8.txt
            └── sub3/
                └── ssub3/
                    └── file9.txt
            """
        ).strip()
    )

    assert (
        test.tree("--del", "--del").strip()
        == dedent(
            """
            /
            └── sub1/
                └── ssub1/
                    └── file5.txt (DEL)
            """
        ).strip()
    )

    assert (
        test.tree("--after", "u2").strip()
        == dedent(
            """
            /
            └── sub3/
                └── ssub3/
                    └── file9.txt
            """
        ).strip()
    )

    assert (
        test.tree("--after", "u2", "--del").strip()
        == dedent(
            """
            /
            ├── sub1/
            │   └── ssub1/
            │       └── file5.txt (DEL)
            └── sub3/
                └── ssub3/
                    └── file9.txt
        """
        ).strip()
    )

    ## Now test recursive even though it's included above
    # print('\n'.join(f"{i!r}," for i in sorted(items)))
    out = test.ls("--no-header")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "file1.txt",
        "file2.txt",
        "sub1/",
        "sub2/",
        "sub3/",
    }

    out = test.ls("--no-header", "--recursive", "--list", "both")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "file1.txt",
        "file2.txt",
        "sub1/",
        "sub1/file3.txt",
        "sub1/file4.txt",
        "sub1/ssub1/",
        "sub1/ssub1/file6.txt",
        "sub2/",
        "sub2/file7.txt",
        "sub2/ssub2/",
        "sub2/ssub2/file8.txt",
        "sub3/",
        "sub3/ssub3/",
        "sub3/ssub3/file9.txt",
    }

    out = test.ls("--no-header", "--recursive", "sub2")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "file7.txt",
        # "ssub2/", # Dirs not included anymore
        "ssub2/file8.txt",
    }

    out = test.ls("--no-header", "--recursive", "sub2", "--full-path", "--list", "both")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "sub2/file7.txt",
        "sub2/ssub2/",
        "sub2/ssub2/file8.txt",
    }

    out = test.ls("--no-header", "--recursive", "sub2", "--list-only", "files")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "file7.txt",
        "ssub2/file8.txt",
    }

    out = test.ls("--no-header", "--recursive", "sub2", "--list-only", "dirs")
    items = {i.strip() for i in out.split("\n") if i.strip()}
    assert items == {
        "ssub2/",
    }


if __name__ == "__main__":
    test_listing()
    # test_del()
    # test_tree()
    print("=" * 50)
    print(" All Passed ".center(50, "="))
    print("=" * 50)



import datetime
import string
import re
import logging

logger = logging.getLogger(__name__)


# This will get monkey-patched later when used in dfb but it is done like this so that
# this module code can be copied to other projects w/o this being affected
def nowfun():
    return datetime.datetime.now().astimezone(datetime.timezone.utc)


def timestamp_parser(timestamp, aware=False, utc=False, epoch=False, now=None):
    """
    Will either accept iso8601 and pass aware,utc, and epoc
    or will take a string that contains one or more of
        ["seconds","minutes","hours","days","weeks"]
    and be a differences from "now" if specified else will use current time

    Inputs:
    -------
    timestamp
        Timestamp to parse. See above for requirements

    aware [False] {True, False, 'utc'}
        If True, will assume timestamps without specified timezones are in
        LOCAL time. If set to 'utc', will assume the unsepcified time zone IS utc time.

        If a timezone is in the timestamp, this will have no effect

    utc [False]
        Convert all responses to UTC time regardless of specified times. If the
        timestamp doesn't have a timezone, it assumed local unless `aware = 'utc'`

    epoch [False]
         Return epoch time instead of a datetime object

    now [None]:
        Set a current time. Defaults to the *actual* current time

    """
    delta = timedelta_parser(timestamp)
    if delta:
        now = now or nowfun()
        timestamp = timestamp_parser(now, aware=aware, utc=utc) - delta

    return iso8601_parser(timestamp, aware=aware, utc=utc, epoch=epoch)


def timedelta_parser(deltastr):
    """
    Return a timedelta object or None
    """
    if not isinstance(deltastr, str):
        return

    deltastr0 = deltastr
    deltastr = deltastr.lower().replace(",", " ")

    keys = ["seconds", "minutes", "hours", "days", "weeks"]

    delta = {}
    for key in keys:
        if key[:-1] not in deltastr:
            continue  # the [:-1] removes an "s"

        val = re.search(
            r"([\d|\.]+)\ *?KEY?".replace("KEY", key),
            deltastr,
            flags=re.IGNORECASE,
        )
        if val:
            delta[key] = float(val.group(1))
    if delta:
        return datetime.timedelta(**delta)
    else:
        return


def iso8601_parser(timestamp, aware=False, utc=False, epoch=False):
    """
    This will parse an ISO-8601 style datetimes including RFC 3339. It is designed to
    parse different variants of YYYY-MM-DD HH:MM:SS, optionally with time zones and
    precision. While it will accept nanosecond precision, it will not capture beyond
    microseconds due to Python's limitations.

    Alternatively, if timestamp is numeric or "i<num>" or "u<num>", it will
    be considered the UNIX Epoch time.

    Only accepts time zone offsets, not names. Years MUST be four digits. Can also accept
    just a date without time (e.g. 2022-06-25) but still must be four digit years.
    Can also accept times without minutes or seconds

    Inputs:
    -------
    timestamp
        Timestamp to parse. See above for requirements

    aware [False] {True, False, 'utc'}
        If True, will assume timestamps without specified timezones are in
        LOCAL time. If set to 'utc', will assume the unsepcified time zone IS utc time.

        If a timezone is in the timestamp, this will have no effect

    utc [False]
        Convert all responses to UTC time regardless of specified times. If the
        timestamp doesn't have a timezone, it assumed local unless `aware = 'utc'`

     epoch [False]
         Return epoch time instead of a datetime object

    Returns:
    --------
    datetime object or epoch float
    """
    timestamp0 = timestamp
    if isinstance(timestamp, str):
        timestamp = timestamp.lower()
        if match := re.match(r"^[i|u](-?[\d|\.]+)$", timestamp):
            timestamp = float(match.group(1))  # May have to deal with 2038 problem?
        elif timestamp.lower().strip() == "now":
            timestamp = datetime.datetime.now().astimezone()

    if isinstance(timestamp, (int, float)):
        timestamp = datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)

    # This is the main formatting. If not a datetime, it is a recursive call to this
    # part below.
    if isinstance(timestamp, datetime.datetime):
        dt = timestamp

        # https://stackoverflow.com/a/27596917 and
        # https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive
        isaware = dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None
        if (aware or utc) and not isaware:
            if isinstance(aware, str) and aware.lower() == "utc":
                dt = dt.replace(tzinfo=datetime.timezone.utc)
            else:  # local
                dt = dt.astimezone()

        if utc:
            dt = dt.astimezone(datetime.timezone.utc)

        if epoch:
            return datetime.datetime.timestamp(dt)
        return dt

    timestamp = timestamp.strip()
    if not timestamp:
        return

    # Handle no time specified but still must be four digit year
    n = sum(c in string.digits for c in timestamp)
    if n <= 6:  # This won't catch them all but still will catch some.
        raise ValueError(
            "MUST at least a FOUR digit year, two digit month, and two digit day. "
            f"Specified: {timestamp0!r}"
        )
    if n == 8:
        timestamp = f"{timestamp} 00:00:00"

    timestamp = timestamp.lower().replace(":", "").replace("t", "").replace("_", "")

    # pull timezone
    if timestamp.endswith("z"):
        tz = "+0000"
        timestamp = timestamp[:-1]
    elif timestamp[-5] in "-+":
        tz = timestamp[-5:]
        timestamp = timestamp[:-5]
    elif timestamp[-3] in "-+":
        tz = timestamp[-3:] + "00"
        timestamp = timestamp[:-3]
    else:
        tz = None

    # Now get rid of anything that isn't numeric or dot. We need to also
    # be weary of something that really just isn't a timestamp.
    l0 = len(timestamp)
    timestamp = "".join(c for c in timestamp if c in string.digits + ".")
    l1 = len(timestamp)

    if l1 / l1 < 0.9:
        logger.debug("Timestamp below tol for bad chars")
        return

    timestamp = timestamp.split(".")
    if len(timestamp) == 1:  # No precision
        timestamp = timestamp[0].ljust(14, "0")  # Pad if not given with MM or SS
        us = ""
    else:
        us = timestamp[-1]
        timestamp = "".join(timestamp[:-1])  # also remove '.'

    # Force microseconds so that we are consistent
    us = round(float(f".{us}000000") * 1e6)
    timestamp = f"{timestamp}.{us:06d}"

    if tz:  # Already will be aware
        res = datetime.datetime.strptime(timestamp + tz, "%Y%m%d%H%M%S.%f%z")
    else:
        res = datetime.datetime.strptime(timestamp, "%Y%m%d%H%M%S.%f")

    # Recursive call to the top for formatting
    return timestamp_parser(res, aware=aware, utc=utc, epoch=epoch)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os, sys, shutil, subprocess
from textwrap import dedent
import hashlib, io

# 3rd Party
import pytest

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

from dfb import rclonecli
from dfb.rclonecli import RcloneCLI

# assert os.path.relpath(rclonecli.__file__, ".") == "rclone.py"  # Must be local!


def rmdir(path):
    try:
        shutil.rmtree(path)
    except OSError:
        pass


def write_config(name):
    try:
        os.makedirs(f"testdirs")
    except:
        pass
    with open(f"testdirs/{name}.cfg", "wt") as fp:
        fp.write(
            dedent(
                f"""\
            [myremote]
            type = alias
            remote = testdirs/{name}"""
            )
        )
    return f"testdirs/{name}.cfg"


def test_main():
    # if True:
    rmdir("testdirs/main")
    cfg = write_config("main")
    rclone = RcloneCLI(
        "myremote:",
        universal_flags=["--config", cfg],
        universal_env={"RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV},
    )

    #####################
    ## Uploads
    #####################

    with open("testdirs/tmpfile.txt", "wt") as fp:
        fp.write("Test File")

    rclone.upload("testdirs/tmpfile.txt")
    assert rclone.read("tmpfile.txt").decode() == "Test File"

    rclone.uploadto("testdirs/tmpfile.txt", "sub/dir/file.txt")
    assert rclone.read("sub/dir/file.txt").decode() == "Test File"

    with open("testdirs/tmpfile2.txt", "wt") as fp:
        fp.write("Test File2")

    rclone.uploadmany(
        [
            "tmpfile.txt",
            "tmpfile2.txt",
        ],
        destdir="many",
        upload_from="testdirs",
    )
    assert rclone.read("many/tmpfile.txt").decode() == "Test File"
    assert rclone.read("many/tmpfile2.txt").decode() == "Test File2"

    # Stream Upload takes four forms: (1) string, (2) bytes, (3) bytes file object
    # (3.5) subprocess.PIPE, (4) iterator/generator. And it needs to be tested
    # with and without fallback mode
    for fallback in [True, False, "auto"]:
        rmdir("testdirs/main/stream")

        rclone.write(b"\x00bytes\x00", "stream/bytes.txt", fallback=fallback)
        rclone.write("strings", "stream/strings.txt", fallback=fallback)
        with open("testdirs/tmpfile.txt", "rb") as fp:
            rclone.write(fp, "stream/file.txt", fallback=fallback)

        proc = subprocess.Popen(["echo", "proc"], stdout=subprocess.PIPE)
        rclone.write(proc.stdout, "stream/proc.txt", fallback=fallback)

        rclone.write(
            (f"line {i}\n".encode() for i in range(10)),
            "stream/generator.txt",
            fallback=fallback,
        )

        files = list(rclone.listremote(subdir="stream"))
        assert {(file["Path"], file["Size"]) for file in files} == {
            ("bytes.txt", 7),
            ("file.txt", 9),
            ("generator.txt", 70),
            ("proc.txt", 5),
            ("strings.txt", 7),
        }

    #####################
    ## Delete
    #####################

    list(rclone.listremote(subdir="tmpfile.txt"))[0]
    rclone.delete("tmpfile.txt", rmdirs=False)
    try:
        list(rclone.listremote(subdir="tmpfile.txt"))[0]
        assert False
    except subprocess.CalledProcessError:
        pass
    assert len(list(rclone.listremote(subdir="many"))) == 2
    rclone.delete("many")
    assert len(list(rclone.listremote(subdir="many"))) == 0

    #####################
    ## Download
    #####################
    rclone.download("stream/file.txt", "testdirs")
    assert open("testdirs/file.txt").read() == "Test File"
    rclone.downloadto("stream/file.txt", "testdirs/afile.txt")
    assert open("testdirs/afile.txt").read() == "Test File"

    #####################
    ## File Object
    #####################

    # Streamupload
    file1 = rclone / "sub" / "1/2" / "file.txt"
    file1.write("uploaded text")
    assert open("testdirs/main/sub/1/2/file.txt").read() == "uploaded text"

    # Upload. Can be local or a remote
    file2 = rclone / "file.txt"
    file2.upload("testdirs/file.txt")
    file2.copy("subdir")
    file2.copyto("subdir/file copy.txt")
    file2.move("moveddir")  # Will *also* change where this is pointed!
    assert file2.remoteitem == "moveddir/file.txt"
    file2.moveto("new/dir/moved.txt")
    assert file2.remoteitem == "new/dir/moved.txt"
    file2.info()

    ##############################
    ## Backend and cached property
    ##############################
    # WARNING: This may break with future rclone versions
    with rclone.capture() as cap:
        assert set(rclone.config_paths) == {"Cache dir", "Config file", "Temp dir"}

        assert rclone.backend_features["Name"] == "local"

        # This should *not* cause another call
        assert rclone.features["About"] == True
        assert rclone.backend_features["Features"]["About"] == True

        assert len(cap.command_history) == 2


def test_fobj():
    rmdir("testdirs/fobj")
    cfg = write_config("fobj")
    rclone = RcloneCLI(
        "myremote:",
        universal_flags=["--config", cfg],
        universal_env={"RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV},
    )

    rclone.write((os.urandom(1024 * 1024) for _ in range(10)), "random.bin")

    # Test with hashes...
    # Make sure we cross boundaries at odd ways by using prime numvers
    with (
        open("testdirs/fobj/random.bin", "rb") as fp1,
        rclone.open(
            "random.bin",
            "rb",
            buffer_size=1000159,  # 976.718 kb, prime number
        ) as fp2,
    ):
        h1 = hashlib.md5()
        h2 = hashlib.md5()

        while b1 := fp1.read(995611):  # 972.276 kb, prime
            h1.update(b1)
        while b2 := fp2.read(995611):
            h2.update(b2)

        assert h1.digest() == h2.digest()

        fp1.seek(1153)
        fp2.seek(1153)
        h1.update(fp1.read(1001669))
        h2.update(fp2.read(1001669))

        assert h1.digest() == h2.digest()

        fp1.seek(1024 * 1024 * 10)
        fp2.seek(1024 * 1024 * 10)

        h1.update(fp1.read())
        h2.update(fp2.read())

        assert h1.digest() == h2.digest()

    rclone.write("lîné1\nlÚne2", "text.txt")
    with rclone.open("text.txt", mode="rt") as fp:
        assert fp.read() == "lîné1\nlÚne2"

    try:
        rclone.open("text.txt", mode="wt")
        assert False
    except io.UnsupportedOperation:
        pass


def test_streamed_output():
    rmdir("testdirs/streamout")
    cfg = write_config("streamout")
    rclone = RcloneCLI(
        "myremote:",
        universal_flags=["--config", cfg],
        universal_env={"RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV},
    )
    lines = rclone.write(
        os.urandom(10),
        "small.bin",
        flags=[
            "--bwlimit",
            "5b",
            "--stats",
            "0.1",
            "--stats-one-line",
            "-v",
        ],
        callopts=dict(stream=True),
    )
    for st, line in lines:
        print(st, line.decode(), end="", flush=True)


def test_context_managers():
    rmdir("testdirs/context")
    cfg = write_config("context")
    rclone = RcloneCLI(
        "myremote:",
        universal_flags=["--config", cfg],
        universal_env={
            "RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV,
            "TEST_ENV": "Testi ng",
        },
    )

    with rclone.capture() as cap:
        rclone.write("test data", "test.txt")
    assert cap.command_history[0] == [
        "rclone",
        "--config",
        "testdirs/context.cfg",
        "rcat",
        "myremote:test.txt",
        "--no-traverse",
        "--no-check-dest",
        "--size",
        "9",
    ]

    with rclone.capture(save_results=True) as cap:
        print(rclone.read("test.txt").decode())
    assert cap.command_history[0] == [
        "rclone",
        "--config",
        "testdirs/context.cfg",
        "cat",
        "myremote:test.txt",
    ]
    assert cap.command_results[0][0] == b"test data"

    # Should do nothing
    with rclone.capture(execute=False) as cap:
        assert rclone.read("test.txt") == None

    # Make sure this works
    with rclone.capture(execute=True) as cap:
        list(rclone.listremote())

    # This should cause a downstream failure
    try:
        with rclone.capture(execute=False) as cap:
            list(rclone.listremote())
        assert False
    except TypeError:
        assert True

    # test the not always
    with open("testdirs/tmpfile.txt", "wt") as fp:
        fp.write("Test File")

    with rclone.capture(save_results=True, save_stdin=True) as cap:
        not_there = "tmpfile.txt: Need to transfer - File not found at Destination"
        unchanged = "tmpfile.txt: Unchanged skipping"
        uncond = "tmpfile.txt: Transferring unconditionally as --ignore-times is in use"
        # First one
        rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there in cap.command_results[-1][1].decode()
        assert unchanged not in cap.command_results[-1][1].decode()
        assert uncond not in cap.command_results[-1][1].decode()

        assert "--no-check-dest" in cap.command_history[-1]
        assert "--ignore-times" not in cap.command_history[-1]

        # Still not "there" because it can't check
        rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there in cap.command_results[-1][1].decode()
        assert unchanged not in cap.command_results[-1][1].decode()
        assert uncond not in cap.command_results[-1][1].decode()

        assert "--no-check-dest" in cap.command_history[-1]
        assert "--ignore-times" not in cap.command_history[-1]

        # Now it shouldn't upload
        with rclone.not_always:
            rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there not in cap.command_results[-1][1].decode()
        assert unchanged in cap.command_results[-1][1].decode()
        assert uncond not in cap.command_results[-1][1].decode()

        assert "--no-check-dest" not in cap.command_history[-1]
        assert "--ignore-times" not in cap.command_history[-1]

        # Test with uploads
        rclone.write("some data", "testdirs/here.txt")  # Should be fine
        rclone.write(b"\xb9\xd5\x85\x85j\xa6-7\xb4\xe3", "testdirs/here.txt")
        rclone.write(io.BytesIO(os.urandom(10)), "testdirs/here.txt")

        shell = cap.shell_script()
        assert "unset RCLONE_PASSWORD_COMMAND" in shell
        assert "echo" in shell
        assert "Could not understand stdin" in shell
        assert "stdin specified and not bytes" in shell

    # Repeat the above with `ignore_times`
    rclone = RcloneCLI(
        "myremote:",
        universal_flags=["--config", cfg],
        universal_env={"RCLONE_PASSWORD_COMMAND": RcloneCLI.DELENV},
        no_check_dest=False,
    )
    rclone.delete("tmpfile.txt")

    with rclone.capture(save_results=True) as cap:
        # First one
        rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there in cap.command_results[-1][1].decode()
        assert unchanged not in cap.command_results[-1][1].decode()
        assert uncond not in cap.command_results[-1][1].decode()

        assert "--no-check-dest" not in cap.command_history[-1]
        assert "--ignore-times" in cap.command_history[-1]

        # Try again. DIFFERENT from above
        rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there not in cap.command_results[-1][1].decode()
        assert unchanged not in cap.command_results[-1][1].decode()
        assert uncond in cap.command_results[-1][1].decode()

        assert "--no-check-dest" not in cap.command_history[-1]
        assert "--ignore-times" in cap.command_history[-1]

        # Now it shouldn't upload
        with rclone.not_always:
            rclone.upload("testdirs/tmpfile.txt", flags="-vv")
        assert not_there not in cap.command_results[-1][1].decode()
        assert unchanged in cap.command_results[-1][1].decode()
        assert uncond not in cap.command_results[-1][1].decode()

        assert "--no-check-dest" not in cap.command_history[-1]
        assert "--ignore-times" not in cap.command_history[-1]


if __name__ == "__main__":
    test_main()
    test_fobj()
    test_streamed_output()
    test_context_managers()

    print("-" * 50)
    print("-- PASSED --")
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os, sys, json
from pathlib import Path
from collections import defaultdict
from math import inf
import logging

from dfb.timestamps import timestamp_parser
from dfb.dstdb import rpath2apath as _rpath2apath, NoTimestampInNameError

_r = repr


def _rs(a):
    return repr(str(a))


def rpath2apath(rpath):
    try:
        return _rpath2apath(rpath)
    except (NoTimestampInNameError, ValueError):
        return rpath, None, ""


class DestNotEmptyError(OSError):
    pass


class FileExistsError(DestNotEmptyError):
    pass


def dfblink(
    mount,
    dest,
    before=None,
    after=None,
    maxdepth=None,
    allow_non_empty=False,
    force_overwrite=False,
):
    """
    Build symlinks from 'dest' to rclone 'mount' location.

    All symlinks will be made to be absolute regardless of specification

    Inputs:
    -------
    mount
        Location of rclone mount of the backup. See note below on subdirectory

    dest
        Location to build the links

    before, after [+Inf,-Inf]
        Timestamps for when to show files. Generally speaking 'before' is the
        time at which to create the links. 'after' can be used to isolate specific
        ranges. 'before' defaults to the latest (+Inf) and after defaults to the
        oldest (-Inf)

    maxdepth [None]
        How far to traverse. Default (None) is no limit. The first directory is 0.

    allow_non_empty [False]
        If True, 'dest' must be empty. Otherwise, will fail.

    force_overwrite [False]
        If True, overwrite symlinks that already exists. Otherwise will fail.

    Subdirs Note
    ------------
    The 'mount' location can point to a sub directory of the backup but the rclone
    mount should be at the highest level. This is to allow for reference files to
    correctly resolve.

    For example, if you backups to 'backups:' and want to look at the subdirectory
    'Documents/Pictures', you should

        $ rclone mount backups: mountpoint [flags ...]

    Then

        >>> dfblink("mountpoint/Documents/Pictures",destdir, ...)

    so that references point above 'Documents/Pictures' will resolve.


    """

    if before is None:
        before = inf
    else:
        before = timestamp_parser(before, epoch=True)

    if after is None:
        after = -inf
    else:
        after = timestamp_parser(after, epoch=True)

    logging.logger.debug(f"{before = }, {after = }")

    mount = Path(mount).resolve()
    dest = Path(dest).resolve()
    logging.logger.debug(f"mount = {_rs(mount)}, dest = {_rs(dest)}")

    dest.mkdir(exist_ok=True, parents=True)
    if not allow_non_empty and any(dest.iterdir()):
        raise DestNotEmptyError(f"{_rs(dest)} is not empty. Set allow_non_empty")

    for root, dirs, rpaths in os.walk(mount):
        rel = os.path.relpath(root, mount)
        depth = rel.count("/") + 1 if rel != "." else 0
        if maxdepth is not None and depth > maxdepth:
            logging.logger.debug("max depth hit")
            del dirs[:]  # Do not go deeper
            continue

        dirs[:] = [d for d in dirs if d != ".dfb"]

        # Group by apaths and filter for before
        apaths = defaultdict(list)
        for rpath in rpaths:
            apath, ts, flag = rpath2apath(rpath)

            # filter. Notice both are inclusive.
            if ts is None:
                logging.info(f"file {_rs(rpath)} is missing a timestamp")
                ts = after  # always keep it but give it the lowest value

            if ts > before:
                logging.logger.debug(f"File {_rs(rpath)} is too new. Skipped")
                continue
            if ts < after:
                logging.logger.debug(f"File {_rs(rpath)} is too old. Skipped")
                continue

            apaths[apath].append((ts, rpath, flag))

        for apath, items in apaths.items():
            items.sort()  # Will use ts, the first tuple arg
            ts, rpath, flag = items[-1]

            if flag == "D":
                continue

            src = mount / rel / rpath
            dst = dest / rel / apath

            if flag == "R":
                src0, dst0 = src, dst
                referent = src.read_text().strip()
                try:
                    referent = json.loads(referent)
                except json.JSONDecodeError:
                    referent = {"ver": 1, "path": referent}

                if referent["ver"] == 1:
                    # Take a guess!

                    src = mount / referent["path"]
                    dst = dst.with_suffix(f".WARNING-V1_Ref{dst.suffix}")

                    logging.warn(
                        "Cannot definitively resolve V1 ref for "
                        f"{_rs(src0.relative_to(mount))}. "
                        f"Guessing {_rs(src.relative_to(mount))} "
                        f"which does{'' if src.exists() else ' NOT'} exist. "
                        f"Changing name to {_rs(dst.relative_to(dest))}"
                    )
                elif referent["ver"] == 2:
                    # drop back to os.path for normalization
                    src = Path(os.path.normpath(src.parent / referent["rel"]))
                else:
                    raise ValueError("Unrecognized ref format")

            if dst.exists():
                logging.logger.debug(f"{_rs(dst)} exists")
                if force_overwrite:
                    dst.unlink()
                else:
                    raise FileExistsError(f"{_rs(dst)} exists. Set force_overwrite")

            logging.info(
                f"Linking {_rs(dst.relative_to(dest))} --> {_rs(src.relative_to(mount))}"
            )
            dst.parent.mkdir(exist_ok=True, parents=True)
            os.symlink(src, dst)



import sys
import logging
import json

from . import apath2rpath, rpath2apath, nowfun
from .utils import timestamp_parser

logger = logging.getLogger(__name__)


def cli_apath2rpath(cliconfig):
    apaths = cliconfig.files

    # This is tested manually
    if apaths.count("-") == 1:  # pragma: no cover
        stdin = sys.stdin.buffer.read()
        stdin = stdin.replace(b"\x00", b"\n")
        stdin = stdin.decode().split("\n")
        stdin = [i for i in stdin if i]

        ix = apaths.index("-")
        apaths = apaths[:ix] + stdin + apaths[ix + 1 :]
    elif apaths.count("-") > 1:
        logger.error("Cannot specify '-' more than once")
        sys.exit(2)

    if cliconfig.date:
        ts = timestamp_parser(cliconfig.date, aware=True)
    else:
        ts = nowfun().obj
    rpaths = [apath2rpath(apath, ts=ts) for apath in apaths]

    sep = b"\x00" if cliconfig.print0 else b"\n"
    out = sep.join(rpath.encode() for rpath in rpaths)

    sys.stdout.buffer.write(out)
    if sys.stdout.isatty():
        sys.stdout.buffer.write(b"\n")


def cli_rpath2apath(cliconfig):
    rpaths = cliconfig.files

    # This is tested manually
    if rpaths.count("-") == 1:  # pragma: no cover
        stdin = sys.stdin.buffer.read()
        stdin = stdin.replace(b"\x00", b"\n")
        stdin = stdin.decode().split("\n")
        stdin = [i for i in stdin if i]

        ix = rpaths.index("-")
        rpaths = rpaths[:ix] + stdin + rpaths[ix + 1 :]
    elif rpaths.count("-") > 1:
        logger.error("Cannot specify '-' more than once")
        sys.exit(2)

    for rpath in rpaths:
        apath, ts, flag = rpath2apath(rpath)
        date = timestamp_parser(ts)
        if cliconfig.timestamp_local:  # pragma: no cover
            date = (
                date.astimezone()
            )  # make it local. Hard to test without knowing timezone

        datestr = date.isoformat()
        res = {"apath": apath, "timestamp": datestr, "flag": flag}

        print(json.dumps(res, indent=None, separators=(",", ":")), flush=True)



"""
Utilities
"""

import os, sys
import datetime
import sqlite3
import random
import subprocess
import shlex
import mimetypes
import re
import logging
from collections import namedtuple

from .timestamps import timestamp_parser, iso8601_parser

logger = logging.getLogger(__name__)

tsrep = namedtuple("timestamps", ("ts", "dt", "obj", "pretty"))


class NoTimestampInNameError(ValueError):
    pass


def time2all(dt_or_ts):
    """Convert from dt or ts to all formats"""
    if isinstance(dt_or_ts, str):
        n = len(dt_or_ts)
        if n == 14:  # No timezone. Assume UTC
            dt_or_ts += "Z"
        elif n < 11:
            try:
                dt_or_ts = int(dt_or_ts)
            except:
                pass
    obj = timestamp_parser(dt_or_ts, utc=True)
    dt = obj.astimezone(datetime.timezone.utc).strftime("%Y%m%d%H%M%S")
    ts = int(obj.timestamp())
    pretty = obj.astimezone().isoformat()
    return tsrep(ts, dt, obj, pretty)


class MyRow(sqlite3.Row):
    """Fancier but performant sqlite3 row"""

    def todict(self):
        return {k: self[k] for k in self.keys()}

    def values(self):
        for k in self.keys():
            yield self[k]

    def items(self):
        for k in self.keys():
            yield k, self[k]

    def get(self, key, default=None):
        try:
            return self[key]
        except:
            return default

    def __str__(self):
        return "Row(" + str(self.todict()) + ")"

    __repr__ = __str__


class star:
    """
    Wrapper to make an iterable of arguments be applied as positional.

        star(fun)(args) --> fun(*args)

    Also passes additional args

        star(fun)(args,other1,other2,kw=value) --> fun(*args,other1,other2,kw=value)
    """

    def __init__(self, fun):
        self.fun = fun

    def __call__(self, arg, *args, **kwargs):
        return self.fun(*arg, *args, **kwargs)


def tabulate(table, indent=2, sep="  "):
    """Fancy printing of data"""
    tabulated = []
    nc = [len(c) for c in table[0]]
    for row in table[1:]:
        for ic, c in enumerate(row):
            nc[ic] = max(nc[ic], len(c))

    for row in table:
        r = [f"{c:>{n}s}" for c, n in zip(row[:-1], nc[:-1])]
        r.append(f"{row[-1]:<{nc[-1]}s}")
        r = " " * indent + sep.join(r).rstrip()
        tabulated.append(r)
    return "\n".join(tabulated)


def human_readable_bytes(
    byte_count,
    base=int(os.environ.get("DFB_BASE", 1024)),  # undocumented environment setting
    short=True,
    fmt=False,
):
    """
    Return a value,label tuple with human readable sizes or,
    if fmt = True, return a formatted version as `{value:0.2f} {label:s}`

    |        Decimal            |            Binary       |
    | 1000    | kB | kilobyte   | 1024     KiB | kibibyte |
    | 1000^2  | MB | megabyte   | 1024^2   MiB | mebibyte |
    | 1000^3  | GB | gigabyte   | 1024^3   GiB | gibibyte |
    | 1000^4  | TB | terabyte   | 1024^4   TiB | tebibyte |
    | 1000^5  | PB | petabyte   | 1024^5   PiB | pebibyte |
    | 1000^6  | EB | exabyte    | 1024^6   EiB | exbibyte |
    | 1000^7  | ZB | zettabyte  | 1024^7   ZiB | zebibyte |
    | 1000^8  | YB | yottabyte  | 1024^8   YiB | yobibyte |

    Example:
        >>> human_readable_bytes(2580417210)
            (2.40320079959929, 'gb')
    """
    if base not in (1024, 1000):
        raise ValueError("base must be 1000 or 1024")

    best = 0
    for ii in range(9):
        if (byte_count / (base**ii * 1.0)) < 1:
            break
        best = ii

    if base == 1000 and short:
        labels = ["B", "kB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    elif base == 1000 and not short:
        labels = ["", "kilo", "mega", "giga", "tera", "peta", "exa", "zetta", "yotta"]
        labels = [l + "byte" for l in labels]
    elif base == 1024 and short:
        labels = ["B", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB"]
    elif base == 1024 and not short:
        labels = ["", "ki", "me", "gi", "te", "pe", "ex", "ze", "yo"]
        labels = [l + "bibyte" for l in labels]

    res = byte_count / (base**best * 1.0), labels[best]
    if fmt:
        return "{0:g} {1:s}".format(*res)
    return res


def parse_bytes(strsize):
    if isinstance(strsize, (int, float)):
        return int(strsize)

    prefix2bytes = {"b": 1}

    dec_prefix = ["", "kilo", "mega", "giga", "tera", "peta", "exa", "zetta", "yotta"]
    bin_prefix = ["", "kibi", "mebi", "gibi", "tebi", "pebi", "exbi", "zebi", "yobi"]

    for ii, (dp, bp) in enumerate(zip(dec_prefix, bin_prefix)):
        prefix2bytes[dp] = 1000**ii
        prefix2bytes[bp] = 1024**ii

        if ii:
            c = dp[0]
            prefix2bytes[c] = prefix2bytes[f"{c}b"] = 1000**ii
            prefix2bytes[f"{c}i"] = prefix2bytes[f"{c}ib"] = 1024**ii

    strsize = (
        strsize.lower()
        .replace("bytes", "")
        .replace("byte", "")
        .replace(" ", "")
        .strip()
    )

    match = re.match(r"([\d|\.]+)(\D*)", strsize)
    if not match:
        raise ValueErorr("Could not parse")
    val, units = match.groups()
    val = float(val)
    try:
        uval = prefix2bytes[units]
    except KeyError:
        raise ValueError(f"Unrecognized {units = }")

    return int(val * uval)


def shell_runner(cmds, dry=False, env=None, prefix=""):
    """
    Run the shell command (string or list) and return the returncode
    """

    environ = os.environ.copy()
    if env:
        environ.update(env)

    kwargs = {}

    prefix = [prefix] or []
    if dry:
        prefix.append("DRY-RUN")

    prefix = ".".join(prefix)

    if isinstance(cmds, str):
        for line in cmds.rstrip().split("\n"):
            logger.info(f"{prefix}: $ {line}")
        shell = True
    elif isinstance(cmds, (list, tuple)):
        logger.info(f"{prefix} {cmds}")
        shell = False
    elif isinstance(cmds, dict):
        logger.info(f"{prefix} {cmds}")
        cmds0 = cmds.copy()
        try:
            cmds = cmds0.pop("cmd")
        except KeyError:
            raise KeyError("Dict shell commands MUST have 'cmd' defined")
        shell = cmds0.pop("shell", False)
        environ.update(cmds0.pop("env", {}))
        cmds0.pop("stdout", None)
        cmds0.pop("stderr", None)
        logger.debug(f"Cleaned cmd: {cmds0}")
        kwargs.update(cmds0)
    else:
        raise TypeError("Shell commands must be str, list/tuple, or dict")

    if dry:
        return logger.info("DRY-RUN: Not running")

    # Apply formatting. Uses the C-Style so that it is less likely to
    # have to need escaping
    if isinstance(cmds, (list, tuple)):
        cmds0 = cmds.copy()
        cmds = [cmd % environ for cmd in cmds]
        if cmds != cmds0:
            logger.debug(f"Formatted cmds: {cmds}")

    proc = subprocess.Popen(
        cmds,
        shell=shell,
        env=environ,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        **kwargs,
    )

    out, err = proc.communicate()
    out, err = out.decode(), err.decode()
    out = out.rstrip("\n")
    err = err.rstrip("\n")

    logger.info(f"out: {out}")

    if err.strip():
        logger.info(f"err: {err}")

    if proc.returncode > 0:
        logger.info(
            f"{prefix} WARNING: Command return non-zero returncode: {proc.returncode}"
        )
    return proc.returncode


def time_format(dt, upper=False):
    """Format time into days (D), hours (H), minutes (M), and seconds (S)"""
    labels = [  # Label, # of sec
        ("D", 60 * 60 * 24),
        ("H", 60 * 60),
        ("M", 60),
        ("S", 1),
    ]
    res = []
    for label, sec in labels:
        val, dt = divmod(dt, sec)
        if not val and not res and label != "S":  # Do not skip if already done
            continue
        if label == "S" and dt > 0:  # Need to handle leftover
            res.append(f"{val+dt:0.2f}")
        elif label in "HMS":  # these get zero padded
            res.append(f"{int(val):02d}")
        else:  # Do not zero pad dats
            res.append(f"{int(val):d}")
        res.append(label if upper else label.lower())
    return "".join(res)


def randstr(N=15):
    c = "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
    return "".join(random.choice(c) for _ in range(N))


def listify(flags):
    """Turn argument into a list. None or False-like become empty list"""
    if isinstance(flags, list):
        return flags
    flags = flags or []
    if isinstance(flags, str):
        flags = [flags]
    return list(flags)


flagify = listify


def dictify(mydict):
    """Turn argument into a dict. None or False-like become empty dict"""
    if isinstance(mydict, dict):
        return mydict
    if not mydict:
        return {}
    if isinstance(mydict, (list, tuple)):
        return dict(mydict)
    return mydict


def shell_header(config, cd=True):
    from .rclonerc import RC

    out = []
    if cd:
        cmd = ["cd", os.path.abspath(os.getcwd())]
        out.append(shlex.join(cmd))

    for key, value in config.rclone_env.items():
        if value in ("**UNSET**", RC.DELENV):
            out.append(f"unset {key}")
            continue
        out.append(f"export {key}={shlex.quote(value)}")

    return "\n".join(out)


def smart_open(filename, mode="rb"):
    filename = str(filename)
    if filename.endswith(".gz"):
        import gzip as gz

        return gz.open(filename, mode)
    elif filename.endswith(".xz"):
        import lzma as xz

        return xz.open(filename, mode)
    else:
        return open(filename, mode)


def head_tail_table(table, /, head=None, tail=None, *, header=True, dots=False):
    """
    head or tail a table.

    Inputs:
    ------
    table
        List of lists represeting a list of rows

    head [None]
        Include the first 'head' rows. If None will not include
        the head of the table unless tail is also None

    tail [None]
        Include the last 'tail' rows. If None, will not include
        the tail of the table unless head is also None

    header [True]
        Exclude the header from 'head' and always return it. If
        False, head will count the first row

    dots [False]
        If true, adds a row with '...' for all cols or a single '...'
        if the rows are not lists. Number of cols is based on the
        first row being a list or tuple and then its length
    """
    if isinstance(table[0], (list, tuple)):
        ncol = len(table[0])
        dotrow = [*["..."] * ncol]
    else:
        ncol = 0
        dotrow = "..."

    head, tail = max([0, head or 0]), max([0, tail or 0])

    if head == tail == 0:
        return table  # Nothing. Even if header

    if not (head and tail):  # no dots if not both
        dots = False

    out = []
    if header:
        out.append(table[0])
        table = table[1:]

    ixhead = set(range(head))
    ixtail = set(range(len(table) - tail, len(table)))
    ixtot = ixhead.union(ixtail)

    headfin = False
    for i, row in enumerate(table):
        # If the full table is covered, this block will never be
        # skipped. But once it is, add the dots the first time then
        # continue. This way you only walk the table once
        if i in ixtot:
            out.append(row)
            continue

        if dots and not headfin and not ixhead.intersection(ixtail):
            out.append(dotrow)
        headfin = True

    return out


def smart_splitext(file):
    """
    Split into stem,ext but allow for multiple valid extensions
    such as .tar.gz.

    Always splits the first extension but keeps going while
    the others are valid MIME types. Never includes the first
    part, even if leading dot
    """
    if not mimetypes.inited:
        mimetypes.init()

    parent, name = os.path.split(file)

    parts = name.split(".")
    if not parts[0]:  # leading dot
        parts[1] = f".{parts[1]}"
        parts = parts[1:]

    if len(parts) == 1:  # Just file.ext
        return file, ""

    # Decide where to stop. This is bounded such that it will
    # always include the first and never include the last
    for ix in range(1, len(parts)):
        if "." + parts[-ix - 1].lower() not in mimetypes.types_map:
            break

    stem = ".".join(parts[:-ix])
    ext = "." + ".".join(parts[-ix:])  # No ext covered above
    return os.path.join(parent, stem), ext


def rpath2apath(rpath):
    """
    Convert the rpath to the apath with the time and flag.

    This is designed to handle a few special cases. Notably, if the flag is
    manually appended on to the file such as if done by hand (incorrectly).

    Also note that in the case of a file like "file.<date1>.<date2>" with
    no extension, it should return "file.<date2>" tagged at <date1> in accordance
    with the split.
    """
    parent, rname = os.path.split(rpath)

    # Case 1: smartsplit off ext. The tag will not be a MIME type
    #         so this will work with file.20220625232247.tar.gz
    #         and file.tar.20220625232247.gz
    # NOTE: This comes FIRST in case of "file.<date1>.<date2>"
    base_w_tag, ext = smart_splitext(rname)
    base, tag = os.path.splitext(base_w_tag)
    try:
        ts, flag = parse_dateflag(tag)
        apath = os.path.join(parent, f"{base}{ext}")
        return apath, ts, flag
    except ValueError:
        pass

    # Case 2: The extension is the end.
    aname, tag = os.path.splitext(rname)
    try:
        ts, flag = parse_dateflag(tag)
        apath = os.path.join(parent, aname)
        return apath, ts, flag
    except ValueError:
        pass

    raise NoTimestampInNameError(f"No timestamp in {rpath = }")


re_datetag = re.compile(
    r"""
    ^                        # Start of the string
    (\d{4})                  # Match any four digits representing the year (YYYY)
    (0[1-9]|1[0-2])          # Match two digits for the month (01-09, 10-12)
    (0[1-9]|[12][0-9]|3[01]) # Match two digits for the day (01-09, 10-29, 30-31)
    ([01][0-9]|2[0-3])       # Match two digits for the hour (00-23)
    ([0-5][0-9])             # Match two digits for the minutes (00-59)
    ([0-5][0-9])             # Match two digits for the seconds (00-59)
    (R|D)?                   # Optionally match an "R" or "D" at the end
    $                        # End of the string
    """,
    re.VERBOSE,
)


def parse_dateflag(ts):
    ts = ts.removeprefix(".")
    if not (match := re_datetag.match(ts)):
        raise ValueError()
    ts, _, _, _ = time2all("".join(match.groups()[:-1]))
    flag = match.groups("")[-1]
    return ts, flag


def apath2rpath(apath, ts=None, *, flag="", verify=True):
    """
    Convert from apath,ts ('sub/dir/file.txt',12345)
    to rpath ('sub/dir/file.12345.txt')

    Will not be correct for references but *will* give the
    referrer path
    """
    from . import nowfun  # Avoid circular import

    ts = ts or nowfun()[0]
    ts, dt, _, _ = time2all(ts)

    base, ext = smart_splitext(apath)
    rpath = f"{base}.{dt}{flag}{ext}"

    # Comment this out b/c older split names will give a false positive.
    # if _verify and rpath2apath(rpath,_verify=False)[0] != apath:
    #     logger.error(
    #         f"Failed round-trip sanity check with {apath = }, {rpath = }. "
    #         "Please submit a bug report"
    #     )

    # Sanity check:
    if verify and rpath2apath(rpath) != (apath, ts, flag):
        logger.warning(
            f"Failed sanity check {apath = }, {rpath = }. Using fallback split"
        )
        base, ext = os.path.split(apath)
        rpath = f"{base}.{dt}{flag}{ext}"
    return rpath



import os, sys
import io
import shutil
from pathlib import Path
import datetime, time
import random
import subprocess
import hashlib
import shutil
import atexit
import logging
from functools import cached_property

logger = logging.getLogger(__name__)

PWD0 = os.path.abspath(os.path.dirname(__file__))
os.chdir(PWD0)

p = os.path.abspath("../")
if p not in sys.path:
    sys.path.insert(0, p)

import dfb.configuration
import dfb.cli
import dfb.dstdb
from dfb.rclonecli import RcloneCLI as RcloneAPI
from dfb.rclonerc import random_port

_r = repr

dfb.cli._TESTMODE = True


class StringBufferBytesIO(io.TextIOWrapper):
    """
    Acts like io.StringIO but also supports .buffer for bytes
    """

    def __init__(self, buffer=None, encoding="utf-8"):
        # If no buffer is provided, create a new BytesIO buffer
        if buffer is None:
            buffer = io.BytesIO()
        super().__init__(buffer, encoding=encoding)

    @property
    def bytes_buffer(self):
        return self.buffer.raw


class Capture:
    def __enter__(self):
        self.old_stdout = sys.stdout
        self.old_stderr = sys.stderr

        self.stdout_buffer = io.BytesIO()
        self.stderr_buffer = io.BytesIO()

        sys.stdout = io.TextIOWrapper(self.stdout_buffer, encoding="utf-8")
        sys.stderr = io.TextIOWrapper(self.stderr_buffer, encoding="utf-8")

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        sys.stdout.flush()
        sys.stderr.flush()
        self.stdout_buffer.flush()
        self.stderr_buffer.flush()

        self.out_bytes = self.stdout_buffer.getvalue()
        self.err_bytes = self.stderr_buffer.getvalue()
        sys.stdout = self.old_stdout
        sys.stderr = self.old_stderr

    @cached_property
    def out(self):
        return self.out_bytes.decode("utf-8")

    @cached_property
    def err(self):
        return self.err_bytes.decode("utf-8")


class Tester:
    def __init__(self, *, name, src=None, dst=None, seed=1):
        os.chdir(PWD0)

        random.seed(seed)

        dfb.configuration._TEMPDIR = "TEMP"
        dfb._override_ts = "1970-01-01 00:00:00Z"  # Offsets = unix time

        self.pwd = Path(os.path.abspath(f"testdirs/{name}"))
        self.make_ignore()

        if src is None:
            src = os.path.join(self.pwd, "src")
        if dst is None:
            dst = os.path.join(self.pwd, "dst")

        self.src = src
        self.dst = dst

        self.config = {
            "src": self.src,
            "dst": self.dst,
            "rclone_env": {
                "RCLONE_CACHE_DIR": str((self.pwd / "cache").resolve()),
                "RCLONE_PASSWORD_COMMAND": RcloneAPI.DELENV,
                "RCLONE_CONFIG": str(self.pwd / "rclone.cfg"),
            },
            "config_id": f"test_{name}",
        }
        self.configfile = str(self.pwd / "config.py")

        try:
            shutil.rmtree(self.pwd)
        except OSError:
            pass

        os.makedirs(self.pwd)
        shutil.copy2("rclone.cfg", self.pwd / "rclone.cfg")

        os.chdir(self.pwd)

        self.logs = []

        self.backup_local_files = {}

        # shutil.copy2("rclone.cfg", self.pwd / "rclone.cfg")
        # self.config["rclone_env"] = {"RCLONE_CONFIG": "rclone.cfg"}

    def write_config(self):
        with open(self.configfile, "wt") as fobj:
            for key, val in self.config.items():
                print(f"{key} = {val!r}", file=fobj)

        self.config_obj = dfb.configuration.Config(self.configfile).parse()
        self.dstdb = dfb.dstdb.DFBDST(self.config_obj)
        return self.config_obj

    def call(self, cmd0, *args, offset=None):
        """Call with a specified config and offset"""
        if offset is None:
            offset = 2 * len(self.logs) + 1
        dfb._override_offset = offset

        r = dfb.cli.cli([cmd0, *args, "--config", self.configfile])

        logfile = self.config_obj.logfile.resolve()
        debugfile = self.config_obj.debuglogfile.resolve()

        tt = time.time_ns() // 1000

        if logfile.exists():
            logtext = logfile.read_text()
            dest = f"{logfile.stem}.{tt}{logfile.suffix}"
            logfile.rename(dest)
        else:
            logtext = ""

        if debugfile.exists():
            debugtext = debugfile.read_text()
            dest = f"{debugfile.stem}.{tt}{debugfile.suffix}"
            debugfile.rename(dest)
        else:
            debugtext = ""

        self.logs.append((logtext, debugtext))
        return r

    def ls(self, *cmd, **kw):
        with Capture() as cap:
            self.call("ls", *cmd, **kw)
        out = cap.out
        print(out)
        return out

    def tree(self, *cmd, **kw):
        with Capture() as cap:
            self.call("tree", *cmd, **kw)
        out = cap.out
        print(out)
        return out

    def backup(self, *args, offset=None, allow_error=False):
        backobj = self.call("backup", *args, offset=offset)
        if not backobj and allow_error:
            return

        self.backup_local_files[offset] = self.local_files()

        return backobj

    def make_ignore(self, file=".ignore"):
        ignore = self.pwd.parent / file
        if not ignore.exists():
            self.pwd.parent.mkdir(exist_ok=True, parents=True)
            with ignore.open(mode="a"):
                pass

    def write(self, path, content, mode="wt", dt=0):
        try:
            os.makedirs(os.path.dirname(path))
        except:
            pass

        with open(path, mode) as file:
            file.write(content)

        if dt:
            change_time(path, dt)

        # Make the times integers to avoid issues
        stat = os.stat(path)
        os.utime(path, (stat.st_atime, stat.st_mtime))

    def write_pre(self, path, content, mode="wt", dt=None):
        """Write items randomly in the past"""
        dt = dt if not None else -5 * (1 + random.random())
        if path.startswith("B"):
            raise ValueError("No pre on B")
        self.write(path, content, mode=mode, dt=dt)

    def write_post(self, path, content, mode="wt", add_dt=0):
        """
        Write items randomly in the future. Can add even more if forcing
        newer
        """
        dt = 5 * (1 + random.random()) + add_dt
        self.write(path, content, mode=mode, dt=dt)

    def read(self, path):
        with open(path, "rt") as file:
            return file.read().strip()

    @staticmethod
    def sha1(path):
        hh = hashlib.sha1()
        with open(path, "rb") as file:
            while dat := file.read(1024 * 512):
                hh.update(dat)
        return hh.hexdigest()

    def globread(self, globpath):
        paths = glob.glob(globpath)
        if len(paths) == 0:
            raise OSError("No files matched the glob pattern")
        if len(paths) > 1:
            raise OSError(f"Too many files matched the pattern: {paths}")

        return self.read(paths[0])

    def move(self, src, dst):
        try:
            os.makedirs(os.path.dirname(dst))
        except OSError:
            pass

        shutil.move(src, dst)

    def remote_snapshot(self, **kwargs):
        """
        Use the snapshot utils to list at a certain time
        """
        files = set()
        for file in map(self.dstdb.fullrow2dict, self.dstdb.snapshot(**kwargs)):
            row = {
                "apath": file["apath"],
                "size": file["size"],
                "mtime": int(file["mtime"]),
                "sha1": self.sha1(os.path.join(self.config_obj.dst, file["rpath"])),
            }
            row = frozenset(row.items())
            files.add(row)
        return files

    def local_files(self, path=None):
        files = set()
        path = path or self.src
        for file in tree(path):
            stat = os.stat(file)
            row = {
                "apath": os.path.relpath(file, path),
                "size": stat.st_size,
                "mtime": int(stat.st_mtime),
                "sha1": Tester.sha1(file),
            }
            row = frozenset(row.items())
            files.add(row)
        return files

    def tree_sha1s(self, path):
        files = set()
        for file in tree(path):
            files.add(Tester.sha1(file))
        return files

    def all_src_in_dst(self, return_diff=False):
        """return if all source files exists *somewhere* in the dest"""
        diff = self.tree_sha1s(self.src) - self.tree_sha1s(self.dst) == set()
        if return_diff:
            return diff
        return diff == set()

    def src_missing_in_dst(self, keys=None, **kwargs):
        diff = self.local_files() - self.remote_snapshot(**kwargs)
        if not keys:
            return diff
        if isinstance(keys, str):
            keys = [keys]
        return {frozenset(i for i in d if i[0] in keys) for d in diff}

    @property
    def src_rclone(self):
        return self.config_obj.src_rclone

    @property
    def dst_rclone(self):
        return self.config_obj.dst_rclone


def change_time(path, time_adj):
    """Change the time on a file path"""
    stat = os.stat(path)
    os.utime(path, (stat.st_atime + time_adj, stat.st_mtime + time_adj))


def tree(path, hidden=False):
    files = []
    for dirpath, dirnames, filenames in os.walk(path, followlinks=True):
        exc = {".DS_Store"}
        filenames = [f for f in filenames if f not in exc]
        if not hidden:
            filenames[:] = [f for f in filenames if not f.startswith(".")]
            dirnames[:] = [d for d in dirnames if not d.startswith(".")]

        filenames.sort(key=str.lower)
        dirnames.sort(key=str.lower)

        files.extend(os.path.join(dirpath, filename) for filename in filenames)

    return files


def dict2frozen(item):
    if isinstance(item, dict):
        return frozenset((k, dict2frozen(v)) for k, v in item.items())
    if isinstance(item, list):
        return tuple(dict2frozen(i) for i in item)
    if isinstance(item, set):
        return frozenset(item)
    return item


def venn(A, B):
    A = set(A)
    B = set(B)
    return {r"A\B": A - B, "A∩B": A.intersection(B), r"B\A": B - A}


class WebDAV:
    def __init__(self, path=None, ip="localhost", port=None):
        self.port = port or random_port()
        self.ip = ip
        self.pwd = path or os.getcwd()
        self.running = False

    @property
    def remote(self):
        if not self.running:
            return
        return f":webdav,url='http://{self.ip}:{self.port}',vendor=rclone:"

    def start(self):
        atexit.register(self.stop)

        cmd = [
            "rclone",
            "serve",
            "webdav",
            self.pwd,
            "--addr",
            f"{self.ip}:{self.port}",
        ]
        logger.debug(f"WEBDAV: {cmd = }")
        self.proc = subprocess.Popen(
            cmd, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL
        )
        self.running = True
        logger.debug("START WEBDAV")
        return self

    def stop(self):
        if not self.running:
            return
        import signal

        self.proc.send_signal(signal.SIGKILL)
        logger.debug("END WEBDAV")
        self.running = False

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, *_, **__):
        self.stop()
